"""
"""
import pygame
from pygame.locals import *

from .const import *
from . import widget


class Input(widget.Widget):
    """A single line text input.

    <pre>Input(value="",size=20)</pre>

    <dl>
    <dt>value<dd>initial text
    <dt>size<dd>size for the text box, in characters
    </dl>

    <strong>Example</strong>
    <code>
    w = Input(value="Cuzco the Goat",size=20)

    w = Input("Marbles")
    </code>

    """

    def __init__(self, value="", size=20, **params):
        params.setdefault("cls", "input")
        widget.Widget.__init__(self, **params)
        self.value = value
        self.pos = len(str(value))
        self.vpos = 0
        self.font = self.style.font
        w, h = self.font.size("e" * size)
        self.style.width, self.style.height = w, h
        # self.rect.w=w+self.style.padding_left+self.style.padding_right;
        # self.rect.h=h+self.style.padding_top+self.style.padding_bottom;

    def paint(self, s):
        r = pygame.Rect(0, 0, self.rect.w, self.rect.h)

        cs = 2  # NOTE: should be in a style

        w, h = self.font.size(self.value[0 : self.pos])
        x = w - self.vpos
        if x < 0:
            self.vpos -= -x
        if x + cs > s.get_width():
            self.vpos += x + cs - s.get_width()

        s.blit(self.font.render(self.value, 1, self.style.color), (-self.vpos, 0))

        if self.container.myfocus is self:
            w, h = self.font.size(self.value[0 : self.pos])
            r.x = w - self.vpos
            r.w = cs
            r.h = h
            s.fill(self.style.color, r)

    def _setvalue(self, v):
        self.__dict__["value"] = v
        self.send(CHANGE)

    def event(self, e):
        if e.type == KEYDOWN:
            if e.key == K_BACKSPACE:
                if self.pos:
                    self._setvalue(self.value[: self.pos - 1] + self.value[self.pos :])
                    self.pos -= 1
            elif e.key == K_DELETE:
                if len(self.value) > self.pos:
                    self._setvalue(self.value[: self.pos] + self.value[self.pos + 1 :])
            elif e.key == K_HOME:
                self.pos = 0
            elif e.key == K_END:
                self.pos = len(self.value)
            elif e.key == K_LEFT:
                if self.pos > 0:
                    self.pos -= 1
            elif e.key == K_RIGHT:
                if self.pos < len(self.value):
                    self.pos += 1
            elif e.key == K_RETURN:
                next(self)
            elif e.key == K_TAB:
                next(self)
            else:
                # c = str(e.unicode)
                c = (e.str).encode("latin-1")
                if c:
                    self._setvalue(self.value[: self.pos] + c + self.value[self.pos :])
                    self.pos += 1
            self.repaint()
        elif e.type == FOCUS:
            self.repaint()
        elif e.type == BLUR:
            self.repaint()

        self.pcls = ""
        if self.container.myfocus is self:
            self.pcls = "focus"

    def __setattr__(self, k, v):
        if k == "value":
            if v == None:
                v = ""
            v = str(v)
            self.pos = len(v)
        _v = self.__dict__.get(k, NOATTR)
        self.__dict__[k] = v
        if k == "value" and _v != NOATTR and _v != v:
            self.send(CHANGE)
            self.repaint()


class Password(Input):
    """A password input, text is *-ed out.

    <pre>Password(value="",size=20)</pre>

    <dl>
    <dt>value<dd>initial text
    <dt>size<dd>size for the text box, in characters
    </dl>

    <strong>Example</strong>
    <code>
    w = Password(value="password",size=20)

    w = Password("53[r3+")
    </code>

    """

    def paint(self, s):
        hidden = "*"
        show = len(self.value) * hidden

        # print "self.value:",self.value

        if self.pos == None:
            self.pos = len(self.value)

        r = pygame.Rect(0, 0, self.rect.w, self.rect.h)

        cs = 2  # NOTE: should be in a style

        w, h = self.font.size(show)
        x = w - self.vpos
        if x < 0:
            self.vpos -= -x
        if x + cs > s.get_width():
            self.vpos += x + cs - s.get_width()

        s.blit(self.font.render(show, 1, self.style.color), (-self.vpos, 0))

        if self.container.myfocus is self:
            # w,h = self.font.size(self.value[0:self.pos])
            w, h = self.font.size(show[0 : self.pos])
            r.x = w - self.vpos
            r.w = cs
            r.h = h
            s.fill(self.style.color, r)



"""
"""
import pygame

from . import container
from . import layout


class _document_widget:
    def __init__(self, w, align=None):
        # w.rect.w,w.rect.h = w.resize()
        # self.rect = w.rect
        self.widget = w
        if align != None:
            self.align = align


class Document(container.Container):
    """A document container contains many widgets strung together in a document format.  (How informative!)

    <pre>Document()</pre>

    """

    def __init__(self, **params):
        params.setdefault("cls", "document")
        container.Container.__init__(self, **params)
        self.layout = layout.Layout(pygame.Rect(0, 0, self.rect.w, self.rect.h))

    def add(self, e, align=None):
        """Add a widget.

        <pre>Document.add(e,align=None)</pre>

        <dl>
        <dt>e<dd>widget
        <dt>align<dd>alignment (None,-1,0,1)
        </dl>
        """
        dw = _document_widget(e, align)
        self.layout.add(dw)
        e.container = self
        e._c_dw = dw
        self.widgets.append(e)
        self.chsize()

    def remove(self, e):
        self.layout._widgets.remove(e._c_dw)
        self.widgets.remove(e)
        self.chsize()

    def block(self, align):
        """Start a new block.

        <pre>Document.block(align)</pre>

        <dl>
        <dt>align<dd>alignment of block (-1,0,1)
        </dl>
        """
        self.layout.add(align)

    def space(self, e):
        """Add a spacer.

        <pre>Document.space(e)</pre>

        <dl>
        <dt>e<dd>a (w,h) size for the spacer
        </dl>
        """
        self.layout.add(e)

    def br(self, height):
        """Add a line break.

        <pre>Document.br(height)</pre>

        <dl>
        <dt>height<dd>height of line break
        </dl>
        """
        self.layout.add((0, height))

    def resize(self, width=None, height=None):
        if self.style.width:
            width = self.style.width
        if self.style.height:
            height = self.style.height

        for w in self.widgets:
            w.rect.w, w.rect.h = w.resize()

            if (width != None and w.rect.w > width) or (
                height != None and w.rect.h > height
            ):
                w.rect.w, w.rect.h = w.resize(width, height)

            dw = w._c_dw
            dw.rect = pygame.Rect(0, 0, w.rect.w, w.rect.h)

        if width == None:
            width = 65535
        self.layout.rect = pygame.Rect(0, 0, width, 0)
        self.layout.resize()

        _max_w = 0

        for w in self.widgets:
            # xt,xl,xb,xr = w.getspacing()
            dw = w._c_dw
            w.style.x, w.style.y, w.rect.w, w.rect.h = (
                dw.rect.x,
                dw.rect.y,
                dw.rect.w,
                dw.rect.h,
            )
            # w.resize()
            w.rect.x, w.rect.y = w.style.x, w.style.y
            _max_w = max(_max_w, w.rect.right)

        # self.rect.w = _max_w #self.layout.rect.w
        # self.rect.h = self.layout.rect.h
        # print 'document',_max_w,self.layout.rect.h
        return _max_w, self.layout.rect.h



"""These widgets are all grouped together because they are non-interactive widgets.
"""

import pygame

from .const import *
from . import widget


class Spacer(widget.Widget):
    """A invisible space.

    <pre>Spacer(width,height)</pre>

    """

    def __init__(self, width, height, **params):
        params.setdefault("focusable", False)
        widget.Widget.__init__(self, width=width, height=height, **params)


class Color(widget.Widget):
    """A block of color.

    <p>The color can be changed at run-time.</p>

    <pre>Color(value=None)</pre>

    <strong>Example</strong>
    <code>
    c = Color()
    c.value = (255,0,0)
    c.value = (0,255,0)
    </code>
    """

    def __init__(self, value=None, **params):
        params.setdefault("focusable", False)
        if value != None:
            params["value"] = value
        widget.Widget.__init__(self, **params)

    def paint(self, s):
        if hasattr(self, "value"):
            s.fill(self.value)

    def __setattr__(self, k, v):
        if k == "value" and type(v) == str:
            v = pygame.Color(v)
        _v = self.__dict__.get(k, NOATTR)
        self.__dict__[k] = v
        if k == "value" and _v != NOATTR and _v != v:
            self.send(CHANGE)
            self.repaint()


class Label(widget.Widget):
    """A text label.

    <pre>Label(value)</pre>

    <dl>
    <dt>value<dd>text to be displayed
    </dl>

    <strong>Example</strong>
    <code>
    w = Label(value="I own a rubber chicken!")

    w = Label("3 rubber chickens")
    </code>
    """

    def __init__(self, value, **params):
        params.setdefault("focusable", False)
        params.setdefault("cls", "label")
        widget.Widget.__init__(self, **params)
        self.value = value
        self.font = self.style.font
        self.style.width, self.style.height = self.font.size(self.value)

    def paint(self, s):
        s.blit(self.font.render(self.value, 1, self.style.color), (0, 0))


class Image(widget.Widget):
    """An image.

    <pre>Image(value)</pre>

    <dl>
    <dt>value<dd>a file name or a pygame.Surface
    </dl>

    """

    def __init__(self, value, **params):
        params.setdefault("focusable", False)
        widget.Widget.__init__(self, **params)
        if type(value) == str:
            value = pygame.image.load(value)
        self.value = value
        self.style.width = self.value.get_width()
        self.style.height = self.value.get_height()

    def paint(self, s):
        s.blit(self.value, (0, 0))



"""Isometric tile engine.

<p>Note -- this engine is not finished, any may not work for your 
particular needs.  If you are able to update it, help would be 
greatly appreciated!</p>

<p>please note that this file is alpha, and is subject to modification in
future versions of pgu!</p>

"""
print("pgu.isovid", "This module is alpha, and is subject to change.")

from pgu.vid import *
import pygame


class Isovid(Vid):
    """Create an iso vid engine.  See [[vid]]"""

    def update(self, screen):
        return self.paint(screen)

    def paint(self, screen):
        sw, sh = screen.get_width(), screen.get_height()

        tlayer = self.tlayer
        blayer = self.blayer
        zlayer = self.zlayer
        w, h = len(tlayer[0]), len(tlayer)

        iso_w, iso_h, iso_z, tile_w, tile_h, base_w, base_h = (
            self.iso_w,
            self.iso_h,
            self.iso_z,
            self.tile_w,
            self.tile_h,
            self.base_w,
            self.base_h,
        )

        base_h2 = base_h / 2
        base_w2 = base_w / 2

        bot = tile_h / base_h2
        todo_max = sh / base_h2 + bot
        todo = [[] for y in range(0, todo_max)]

        self.view.w, self.view.h = sw, sh
        view = self.view
        adj = self.adj = pygame.Rect(-self.view.x, -self.view.y, 0, 0)

        for s in self.sprites:
            self.sprite_calc_irect(s)
            x, y = self.iso_to_view((s.rect.centerx, s.rect.centery))
            v = (y + adj.y) / base_h2 - 1
            if v >= 0 and v < todo_max:
                todo[v].append((s.image, s.irect))
            # else: print 'doesnt fit',v

        w, h = len(tlayer[0]), len(tlayer)
        tiles = self.tiles

        # ""
        if self.bounds == None:
            tmp, y1 = self.tile_to_view((0, 0))
            x1, tmp = self.tile_to_view((0, h + 1))
            tmp, y2 = self.tile_to_view((w + 1, h + 1))
            x2, tmp = self.tile_to_view((w + 1, 0))
            self.bounds = pygame.Rect(x1, y1, x2 - x1, y2 - y1)
        # ""

        if self.bounds != None:
            self.view.clamp_ip(self.bounds)

        ox, oy = self.screen_to_tile((0, 0))
        sx, sy = self.iso_to_view((ox * iso_w, oy * iso_h))
        dx, dy = sx - self.view.x, sy - self.view.y

        for i2 in range(-bot, self.view.h / base_h2 + bot):
            tx, ty = ox + i2 / 2 + i2 % 2, oy + i2 / 2
            x, y = (i2 % 2) * base_w2 + dx, i2 * base_h2 + dy

            # to adjust for the -1 in i1
            x, tx, ty = x - base_w, tx - 1, ty + 1
            for i1 in range(-1, self.view.w / base_w + 2):  # NOTE: not sure why +2
                if ty >= 0 and ty < h and tx >= 0 and tx < w:
                    z = zlayer[ty][tx] * iso_z
                    if blayer != None:
                        n = blayer[ty][tx]
                        if n != 0:
                            t = tiles[n]
                            if t != None and t.image != None:
                                screen.blit(t.image, (x - base_w2, y + z))
                    n = tlayer[ty][tx]
                    if n != 0:
                        t = tiles[n]
                        if t != None and t.image != None:
                            screen.blit(
                                t.image, (x - base_w2, y - (t.image_h - base_h) + z)
                            )

                tx += 1
                ty -= 1
                x += base_w
            for img, irect in todo[y / base_h2]:
                screen.blit(img, (irect.x + adj.x, irect.y + adj.y))

        return [pygame.Rect(0, 0, screen.get_width(), screen.get_height())]

    def iso_to_view(self, pos):
        tlayer = self.tlayer
        w, h = len(tlayer[0]), len(tlayer)

        x, y = pos

        # nx,ny = (h*self.iso_w + x - y)/2, (0 + x + y)/2
        nx, ny = (x - y) / 2, (0 + x + y) / 2

        return (nx * self.base_w / self.iso_w), (ny * self.base_h / self.iso_h)

    def view_to_iso(self, pos):
        tlayer = self.tlayer
        w, h = len(tlayer[0]), len(tlayer)

        x, y = pos

        x, y = x * self.iso_w / self.base_w, y * self.iso_h / self.base_h

        # x -= (self.iso_w/2) * h
        # x -= (self.iso_w/2) * h

        nx = x + y
        ny = y * 2 - nx

        return nx, ny

    def tile_to_view(self, pos):
        return self.iso_to_view((pos[0] * self.iso_w, pos[1] * self.iso_h))

    def screen_to_tile(self, pos):
        x, y = pos
        x += self.view.x
        y += self.view.y
        x, y = self.view_to_iso((x, y))
        return x / self.iso_w, y / self.iso_h

    def tile_to_screen(self, pos):
        x, y = self.iso_to_view((pos[0] * self.iso_w, pos[1] * self.iso_h))
        return x - self.view.x, y - self.view.y

    def tga_load_tiles(self, fname, size, tdata={}):
        Vid.tga_load_tiles(self, fname, size, tdata)

        self.tile_w, self.tile_h = size
        self.iso_w, self.iso_h, self.iso_z = self.tile_w, self.tile_w, 1
        self.base_w, self.base_h = self.tile_w, self.tile_w / 2

    def resize(self, size, bg=0):
        Vid.resize(self, size, bg)

        tlayer = self.tlayer
        w, h = len(tlayer[0]), len(tlayer)

        self.zlayer = [[0 for x in range(0, w)] for y in range(0, h)]

    def sprite_calc_irect(self, s):
        tlayer = self.tlayer
        w, h = len(tlayer[0]), len(tlayer)
        zlayer = self.zlayer

        x, y = self.iso_to_view((s.rect.centerx, s.rect.centery))
        tx, ty = s.rect.centerx / self.iso_w, s.rect.centery / self.iso_h
        z = 0
        if ty >= 0 and ty < h and tx >= 0 and tx < w:
            z = zlayer[ty][tx] * self.iso_z

        nx, ny = x - s.shape.centerx, y - s.shape.centery + z

        s.irect.x, s.irect.y = nx, ny



"""Some handy algorithms for use in games, etc.

<p>please note that this file is alpha, and is subject to modification in
future versions of pgu!</p>
"""
print("pgu.algo", "This module is alpha, and is subject to change.")

# def dist(a,b):
#    return abs(a[0]-b[0]) + abs(a[1]-b[1])


class node:
    def __init__(self, prev, pos, dest):
        self.prev, self.pos, self.dest = prev, pos, dest
        if self.prev == None:
            self.g = 0
        else:
            self.g = self.prev.g + 1
        self.h = dist(pos, dest)
        self.f = self.g + self.h


def astar(start, end, layer, _dist):
    """uses the a* algorithm to find a path

    <pre>astar(start,end,layer,dist): return [list of positions]</pre>

    <dl>
    <dt>start<dd>start position
    <dt>end<dd>end position
    <dt>layer<dd>a grid where zero cells are open and non-zero cells are walls
    <dt>dist<dd>a distance function dist(a,b)
    </dl>

    <p>returns a list of positions from start to end</p>
    """
    global dist
    dist = _dist
    if layer[start[1]][start[0]]:
        return []  # start is blocked
    if layer[end[1]][end[0]]:
        return []  # end is blocked
    w, h = len(layer[0]), len(layer)
    if start[0] < 0 or start[1] < 0 or start[0] >= w or start[1] >= h:
        return []  # start outside of layer
    if end[0] < 0 or end[1] < 0 or end[0] >= w or end[1] >= h:
        return []  # end outside of layer

    opens = []
    open = {}
    closed = {}
    cur = node(None, start, end)
    open[cur.pos] = cur
    opens.append(cur)
    while len(open):
        cur = opens.pop(0)
        if cur.pos not in open:
            continue
        del open[cur.pos]
        closed[cur.pos] = cur
        if cur.pos == end:
            break
        for dx, dy in [
            (0, -1),
            (1, 0),
            (0, 1),
            (-1, 0),
        ]:  # (-1,-1),(1,-1),(-1,1),(1,1)]:
            x, y = pos = cur.pos[0] + dx, cur.pos[1] + dy
            if layer[y][x]:
                continue
            # check for blocks of diagonals
            if layer[cur.pos[1] + dy][cur.pos[0]]:
                continue
            if layer[cur.pos[1]][cur.pos[0] + dx]:
                continue
            new = node(cur, pos, end)
            if pos in open and new.f >= open[pos].f:
                continue
            if pos in closed and new.f >= closed[pos].f:
                continue
            if pos in open:
                del open[pos]
            if pos in closed:
                del closed[pos]
            open[pos] = new
            lo = 0
            hi = len(opens)
            while lo < hi:
                mid = (lo + hi) / 2
                if new.f < opens[mid].f:
                    hi = mid
                else:
                    lo = mid + 1
            opens.insert(lo, new)

    if cur.pos != end:
        return []

    path = []
    while cur.prev != None:
        path.append(cur.pos)
        cur = cur.prev
    path.reverse()
    return path


def getline(a, b):
    """returns a path of points from a to b

    <pre>getline(a,b): return [list of points]</pre>

    <dl>
    <dt>a<dd>starting point
    <dt>b<dd>ending point
    </dl>

    <p>returns a list of points from a to b</p>
    """

    path = []

    x1, y1 = a
    x2, y2 = b
    dx, dy = abs(x2 - x1), abs(y2 - y1)

    if x2 >= x1:
        xi1, xi2 = 1, 1
    else:
        xi1, xi2 = -1, -1

    if y2 >= y1:
        yi1, yi2 = 1, 1
    else:
        yi1, yi2 = -1, -1

    if dx >= dy:
        xi1, yi2 = 0, 0
        d = dx
        n = dx / 2
        a = dy
        p = dx
    else:
        xi2, yi1 = 0, 0
        d = dy
        n = dy / 2
        a = dx
        p = dy

    x, y = x1, y1
    c = 0
    while c <= p:
        path.append((x, y))
        n += a
        if n > d:
            n -= d
            x += xi1
            y += yi1
        x += xi2
        y += yi2
        c += 1
    return path



import pygame
from pygame.locals import *
import random

from .pgu import timer

rr = random.randrange


class Part:
    def __init__(self, pos):
        self.pos = pygame.Rect(pos[0], pos[1], 1, 1)
        self._pos = pygame.Rect(self.pos)
        self.frame = 0


class Effect:
    def __init__(self, total, add):
        if not hasattr(Effect, "images"):
            Effect._init(self)

        self.total = total
        self.add = add

        self.parts = []

        self.frame = 0

    def _init(self):
        Effect.images = images = []
        for r in range(0, 32):
            img = pygame.Surface((r * 2, r * 2)).convert()
            img.fill((255, 0, 255))
            img.set_colorkey((255, 0, 255))
            img.set_alpha(255 - r * 8)

            if r < 8:
                rr = r - 0
                c = (255, 255, 255 - rr * 8)
            elif r < 16:
                rr = r - 8
                c = (255, 255 - rr * 16, 255 - (rr + 8) * 8)
            else:
                rr = r - 16
                c = (255 - rr * 8, 128 - rr * 8, 128 - rr * 8)
            # print c
            pygame.draw.circle(img, c, (r, r), r / 2)
            # pygame.draw.circle(img,(255-r*8,255-r*8,255-r*8),(r,r),r)
            images.append(img)

    def paint(self, screen, origin):
        todo = [[] for n in range(0, 32)]
        for part in self.parts:
            p = part.pos
            r = part.frame
            if r >= 0 and r < 32:
                todo[r].append(p)
        for r in range(31, -1, -1):
            img = Effect.images[r]
            v = 255 - r * 8
            # c = (v,v,v)
            # c = (255,255,255,v)
            for p in todo[r]:
                # pygame.draw.circle(screen,c,(p.x+pos[0],p.y+pos[1]),r+3)
                screen.blit(img, (-origin[0] + p.x - r, -origin[1] + p.y - r))

    def loop(self, pos):
        parts = self.parts
        if len(parts) < self.total:
            for n in range(0, self.add):
                part = Part((pos[0] + 0, pos[1] + rr(-2, 0)))
                parts.append(part)
        for part in parts:
            p, _p = part.pos, part._pos

            dx, dy = p.x - _p.x, p.y - _p.y
            _p.x, _p.y = p.x, p.y

            p.x += dx / 3 + rr(-1, 2)
            p.y += dy / 4 + rr(-2, 1)

            part.frame += 1
            if part.frame >= 32:
                p.x, p.y = pos[0] + 0, pos[1] + rr(-2, 0)
                _p.x, _p.y = p.x, p.y
                part.frame = 0


if __name__ == "__main__":
    screen = pygame.display.set_mode((640, 480))

    ss = [((rr(160, 640), rr(160, 480)), Effect(64, 1)) for n in range(0, 8)]

    t = timer.Timer(40)

    _quit = False
    while not _quit:
        for e in pygame.event.get():
            if e.type == QUIT:
                _quit = True
            if e.type == KEYDOWN and e.key == K_ESCAPE:
                _quit = True

        screen.fill((0, 0, 0))
        for pos, s in ss:
            s.loop(pos)
            s.paint(screen, (0, 0))
        pygame.display.flip()

        t.tick()



from .const import *
from . import widget
from . import app


class ProgressBar(widget.Widget):
    """A progress bar.

    <pre>ProgressBar(value,min,max)</pre>

    <dl>
    <dt>value<dd>starting value
    <dt>min<dd>minimum value rendered on the screen (usually 0)
    <dt>max<dd>maximum value
    </dl>

    <strong>Example</strong>
    <code>
    w = gui.ProgressBar(0,0,100)
    w.value = 25
    </code>
    """

    def __init__(self, value, min, max, **params):
        params.setdefault("cls", "progressbar")
        widget.Widget.__init__(self, **params)
        self.min, self.max, self.value = min, max, value

    def paint(self, s):
        r = pygame.rect.Rect(0, 0, self.rect.w, self.rect.h)
        r.w = r.w * (self.value - self.min) / (self.max - self.min)
        self.bar = r
        app.App.app.theme.render(s, self.style.bar, r)

    def __setattr__(self, k, v):
        if k == "value":
            v = int(v)
            v = max(v, self.min)
            v = min(v, self.max)
        _v = self.__dict__.get(k, NOATTR)
        self.__dict__[k] = v
        if k == "value" and _v != NOATTR and _v != v:
            self.send(CHANGE)
            self.repaint()



import pygame
from pygame.locals import *
import os
import random


import pygame.draw


from . import util
from .pgu import gui

try:
    # import pgu.html
    from . import html
except ImportError:
    # a fake html write function.
    class __html(object):
        def write(self, *args):
            pass

    html = __html()

from .const import *
from . import units

from . import messages

from . import cyclic_list
import pygame.sprite

# class


class StatsDraw:
    """draws the stats health, coal, water, steam on the left side of interface."""

    def __init__(self):
        self.frames = 0
        self.update_frame = {}
        self.update_which = cyclic_list.cyclic_list(
            ["Health", "Coal", "Water", "Steam"]
        )
        self.frames_health = 0

    def draw_img(self, screen, min, max, current, color, where_rect, astat):
        """eg. min - 0, max - 100, current - 45"""

        # this draws an image cutting off the top bit

        height = ((max - min) // (max * 1.0)) * current
        graph_height = where_rect.h
        height = (graph_height // (max * 1.0)) * current

        clip_rect = pygame.Rect(where_rect)
        clip_rect.x = 0
        clip_rect.y = 0
        clip_rect.h = int(height)

        diff = where_rect.h - int(height)
        clip_rect.y += diff

        im_border = None

        if astat == "Health":
            # show the heart every pulse every half a second or so.
            if self.frames_health > (FPS // 2):
                k = "heart.png"
                if self.frames_health > FPS:
                    self.frames_health = 0
            else:
                k = "heart_pulse.png"

            im = self.images[k]

            if k + "laplacian" not in self.images:
                self.images[k + "laplacian"] = pygame.transform.laplacian(im)

            im_border = self.images[k + "laplacian"]

            c = clip_rect.clip(im.get_rect())
            if not c.width and not c.height:
                s = None
            else:
                print((im, c))
                s = im.subsurface(c)

        else:
            raise ValueError("not implemented for this stat")

        draw_rect = pygame.Rect(where_rect)
        draw_rect.y += diff
        if s:
            screen.blit(s, draw_rect)

        if im_border:
            screen.blit(im_border, where_rect)

        return where_rect

    def draw_rect(self, screen, min, max, current, color, where_rect):
        """eg. min - 0, max - 100, current - 45"""
        draw_rect = pygame.Rect(where_rect)

        height = ((max - min) // (max * 1.0)) * current

        graph_height = draw_rect.h

        height = (graph_height // (max * 1.0)) * current

        draw_rect.h = int(height)
        draw_rect.y += where_rect.h - draw_rect.h

        # draw_rect = multr( dxdy, draw_rect)

        s = screen.subsurface(draw_rect)
        s.fill(color)
        return where_rect

    def update_stats(self, screen, stats, robots, max_robots):
        # def update_stats(self, screen, stats):
        # print stats
        # print max_robots

        # return []

        self.interface.dirty["left_background"] = 1
        self.interface.update_left_background(screen)

        stat_rect = {}
        stat_rect["Health"] = S_HEALTH
        stat_rect["Coal"] = S_COAL
        stat_rect["Water"] = S_WATER
        stat_rect["Steam"] = S_STEAM
        stat_rect["CannonPressure"] = S_CANNON_PRESSURE

        stat_color = {}
        stat_color["Health"] = (255, 0, 0)
        stat_color["Coal"] = (0, 0, 0)
        stat_color["Water"] = (0, 0, 255)
        stat_color["Steam"] = (200, 200, 200)
        stat_color["CannonPressure"] = (100, 100, 200)

        updates = []
        update_thisone = self.update_which.nextone()

        for astat in ["Health", "Coal", "Water", "Steam", "CannonPressure"]:
            # update cannon pressure every frame.  Spread the others out.
            if not (astat in [update_thisone, "CannonPressure"]):
                continue

            where_rect = stat_rect[astat]
            color = stat_color[astat]
            min = 0
            max = int(stats["Max" + astat])
            current = int(stats[astat])
            # print where_rect

            # only draw the screen where the rectangle will be drawn.
            if 0:
                c = screen.get_clip()
                screen.set_clip(where_rect)
                screen.blit(self.images["background_status_left.png"], S_STATUS)
                screen.set_clip(c)

            if astat in ["Health"]:
                r = self.draw_img(screen, min, max, current, color, where_rect, astat)
            else:
                r = self.draw_rect(screen, min, max, current, color, where_rect)
            updates.append(r)

        # robots.
        if 1:
            where_rect = S_PEASANTS_REMAINING
            #             color = (200,200,40)
            #             min =0
            #             max = max_robots
            #             current = int(len(robots))
            #
            #             updates = []
            #             r = self.draw_rect(screen, min, max, current, color, where_rect)

            # top right corner remaining 'robots' are shown.

            if self.robots_update:
                r = where_rect
                sc = screen.subsurface(where_rect)
                sc.fill((0, 0, 0))
                img = self.robots_image
                w = sc.get_width() - img.get_width()
                h = sc.get_height() - img.get_height()

                for n in range(0, int(self.robots_total)):
                    sc.blit(img, (random.randrange(0, w), random.randrange(0, h)))

                updates.append(r)

        self.frames_health += 1
        self.frames += 1
        return updates


class Interface:
    def __init__(self):
        """ """

        self.frames = 0
        self.message = messages.generate()
        self.last_message = ""

        self.message_font = pygame.font.Font(data_dir("menu", "vinque.ttf"), 20)
        self.equipment_message_font = pygame.font.Font(
            data_dir("menu", "vinque.ttf"), 14
        )

        self.stats = {}
        self.stats_draw = StatsDraw()
        self.stats_draw.interface = self

        self.equipment_last_message = ""
        self.equipment_message = ""

    def loop(self):
        self.frames += 0

    def load(self):
        """loads the graphics/sounds/resources needed."""

        images_needed = [
            "health_overlay.png",
            "coal_overlay.png",
            "water_overlay.png",
            "steam_overlay.png",
            "background_status_left.png",
            "background_bottom.png",
            "background_illustration.png",
            "background_equipment.png",
            "background_buttons.png",
            "background_messages.png",
            "button_save.png",
            "button_load.png",
            "button_quit.png",
            "button_news.png",
            "button_save_rollover.png",
            "button_load_rollover.png",
            "button_quit_rollover.png",
            "button_news_rollover.png",
            "button_save_down.png",
            "button_load_down.png",
            "button_quit_down.png",
            "button_news_down.png",
            "heart.png",
            "heart_pulse.png",
        ]
        self.images = {}
        self.stats_draw.images = self.images

        for i in images_needed:
            image_path = data_dir("gfx", i)
            print(image_path)
            if os.path.exists(image_path):
                self.images[i] = pygame.image.load(image_path)
                if i in ["heart.png", "heart_pulse.png"]:
                    self.images[i] = self.images[i].convert_alpha()
                else:
                    self.images[i] = self.images[i].convert()
            else:
                self.images[i] = None

        def scale_images(images):
            screen_size = pygame.display.get_surface().get_size()
            # scaler = pygame.transform.smoothscale
            scaler = pygame.transform.scale
            new_images = {}

            orig_size = 640, 480
            for i, surf in list(images.items()):
                if images[i]:
                    r = surf.get_rect()
                    size_scale = util.scale_rect(r, orig_size, screen_size)[2:]

                    new_images[i] = scaler(surf, size_scale)
                else:
                    new_images[i] = surf
            return new_images

        self.images = scale_images(self.images)

        self.bkgrb = pygame.Surface((S_BOTTOM_MESSAGES.w, S_BOTTOM_MESSAGES.h))
        self.bkgrb = self.bkgrb.convert_alpha()
        self.bkgrb.fill((0, 0, 0, 70))

        self.dirty = {}
        d = self.dirty

        d["messages"] = 1
        d["bottom_background"] = 1
        d["left_background"] = 1
        d["background_equipment"] = 1
        d["background_illustration"] = 1
        d["background_buttons"] = 1
        d["background_messages"] = 1
        d["equipment_words"] = 1
        d[""] = 1

    def event(self, g, e):
        pass

        # TODO: handle button clicks.
        if e.type == MOUSEBUTTONDOWN:
            x, y = e.pos
            print((x, y))

            if S_BUTTONS_SAVE.collidepoint(x, y):
                print("save")

            if S_BUTTONS_LOAD.collidepoint(x, y):
                print("load")

            if S_BUTTONS_QUIT.collidepoint(x, y):
                print("quit")

            if S_BUTTONS_NEWS.collidepoint(x, y):
                print("news")
        if e.type == KEYDOWN:
            if e.key == K_m:
                print("ASDFSADF")
                self.new_random_message()

    def new_stats(self, stats):
        """draws the stats with new values."""

    def new_random_message(self):
        self.last_message = self.message
        self.message = messages.generate()
        self.message = self.message.replace("ZANTHOR", "<b>ZANTHOR</b>")
        self.message = self.message.replace("Zanthor", "<b>ZANTHOR</b>")
        self.message = self.message.replace("zanthor", "<b>ZANTHOR</b>")
        self.dirty["messages"] = 1
        self.dirty["background_messages"] = 1

    def new_upgrade_message(self, upgrade_what):
        if upgrade_what:
            self.last_message = self.message
            self.message = messages.generate_upgrade_message(upgrade_what)
        self.dirty["messages"] = 1
        self.dirty["background_messages"] = 1

    def update_messages(self, screen):
        """ """

        # the messages are the same... no need to update.
        if not self.dirty["messages"]:
            if self.message == self.last_message:
                self.dirty["messages"] = 0
                return []

        if self.dirty["messages"]:
            self.dirty["messages"] = 0

            font = self.message_font  # pygame.font.Font(None, 18)
            rect = S_MESSAGES
            rect2 = pygame.Rect(S_MESSAGES)
            rect2.x -= 1
            rect2.y += 1

            aa = 0
            # color = (0x55,0xff,0x55) #(80,80,80)
            color = (0xFF, 0x55, 0x55)  # (80,80,80)
            color2 = (0, 0, 0)  # (80,80,80)

            bgcolor = (0, 0, 0, 0)
            s = screen
            # screen.blit(self.bkgrb, S_MESSAGES)

            # pgu.html.render(font,rect,text,aa,color,bgcolor)
            html.write(s, font, rect2, self.message, aa, color2)
            html.write(s, font, rect, self.message, aa, color)

            return [rect]
        else:
            return []

    def update_bottom_background(self, screen):
        if self.dirty["bottom_background"]:
            self.dirty["bottom_background"] = 0
            screen.blit(self.images["background_bottom.png"], S_BOTTOM)

            return [S_BOTTOM]
        else:
            return []

    def update_left_background(self, screen):
        if self.dirty["left_background"]:
            # FIXME: HACK: TODO: this should be 0 when stats update properly.
            self.dirty["left_background"] = 0
            screen.blit(self.images["background_status_left.png"], S_STATUS)
            return [S_STATUS]
        else:
            return []

    def new_equipment(self, stats):
        self.equipment_last_message = self.equipment_message
        self.equipment_message = "<br><br>"

        for upword in units.upgrade_words:
            self.equipment_message += "%s/%s " % (
                stats["upgrade_amounts"][upword].idx,
                5,
            )
            self.equipment_message += "" + upword.replace("Up", "") + "<br>"
            # TODO: HACK: fix up 5 to be the maximum for that level.

        self.dirty["equipment_words"] = 1
        self.dirty["background_equipment"] = 1

    def update_background_equipment(self, screen):
        if self.dirty["background_equipment"]:
            self.dirty["background_equipment"] = 0
            screen.blit(self.images["background_equipment.png"], S_ITEMS)
            return [S_ITEMS]
        else:
            return []

    def update_equipment_words(self, screen):
        """ """

        # the messages are the same... no need to update.
        if not self.dirty["equipment_words"]:
            if self.equipment_message == self.equipment_last_message:
                self.dirty["equipment_words"] = 0
                return []

        if self.dirty["equipment_words"]:
            self.dirty["equipment_words"] = 0

            font = self.equipment_message_font  # pygame.font.Font(None, 18)
            rect = S_ITEMS
            rect2 = pygame.Rect(S_ITEMS)
            rect2.x -= 1
            rect2.y += 1

            aa = 0
            color = (0x55, 0xFF, 0x55)  # (80,80,80)
            color2 = (0, 0, 0)
            bgcolor = (0, 0, 0, 0)
            s = screen

            # pgu.html.render(font,rect,text,aa,color,bgcolor)
            html.write(s, font, rect2, self.equipment_message, aa, color2)
            html.write(s, font, rect, self.equipment_message, aa, color)

            return [rect]
        else:
            return []

    def update_background_illustration(self, screen):
        if self.dirty["background_illustration"]:
            self.dirty["background_illustration"] = 0
            screen.blit(self.images["background_illustration.png"], S_ROBOT)

            return [S_ROBOT]
        else:
            return []

    def update_background_buttons(self, screen):
        if self.dirty["background_buttons"]:
            self.dirty["background_buttons"] = 0
            s = screen.subsurface(S_BUTTONS)
            surf = self.images["background_bottom.png"]

            screen.blit(surf, S_BUTTONS, S_BOTTOM_BUTTONS)
            return [S_BUTTONS]
        else:
            return []

    def update_background_buttons(self, screen):
        if self.dirty["background_messages"]:
            self.dirty["background_messages"] = 0
            surf = self.images["background_bottom.png"]

            screen.blit(surf, S_MESSAGES, S_BOTTOM_MESSAGES)
            return [S_MESSAGES]
        else:
            return []

    def update(self, tv, screen):
        """returns a list of rects."""

        if self.equipment_message == "":
            self.new_equipment(tv.castle.unit.stats)

        updates = []

        # temporary interface boxes.
        if 0:
            s = screen.subsurface(S_HEALTH)
            s.fill((255, 0, 0))
            updates.append(S_HEALTH)

            s = screen.subsurface(S_COAL)
            s.fill((0, 0, 0))
            updates.append(S_COAL)

            s = screen.subsurface(S_WATER)
            s.fill((0, 0, 255))
            updates.append(S_WATER)

            s = screen.subsurface(S_STEAM)
            s.fill((200, 200, 200))
            updates.append(S_STEAM)

        updates.extend(self.update_bottom_background(screen))
        updates.extend(self.update_left_background(screen))

        updates.extend(self.update_background_buttons(screen))
        updates.extend(self.update_background_illustration(screen))
        updates.extend(self.update_background_equipment(screen))

        updates.extend(self.update_background_buttons(screen))

        # updates.extend( self.update_status_left(screen) )
        # updates.extend( self.update_background_equipment(screen) )
        # updates.extend( self.update_(screen) )

        if 0:
            screen.blit(self.images["background_status_left.png"], S_STATUS)
            updates.append(S_STATUS)

            screen.blit(self.images["background_equipment.png"], S_ITEMS)
            updates.append(S_ITEMS)

            screen.blit(self.images["background_illustration.png"], S_ROBOT)
            updates.append(S_ROBOT)

        # s = screen.subsurface(S_MESSAGES)
        # s.fill((0,200,0))

        m = self.update_messages(screen)
        updates.extend(m)

        ew = self.update_equipment_words(screen)
        updates.extend(ew)

        stats_rects = self.stats_draw.update_stats(
            screen, self.tv.castle.unit.stats, self.tv.robots, self.max_robots
        )
        updates.extend(stats_rects)

        # s = screen.subsurface(S_ROBOT)
        # s.fill((200,200,200))
        # updates.append(S_ROBOT)

        # s = screen.subsurface(S_ITEMS)
        # s.fill((250,250,250))
        # updates.append(S_ITEMS)

        # s = screen.subsurface(S_BUTTONS)
        # s.fill((150,150,200))

        if 0:
            s = screen.subsurface(S_BUTTONS_SAVE)
            s.fill((50, 50, 20))
            s = screen.subsurface(S_BUTTONS_LOAD)
            s.fill((50, 50, 20))
            s = screen.subsurface(S_BUTTONS_QUIT)
            s.fill((50, 50, 200))
            s = screen.subsurface(S_BUTTONS_NEWS)
            s.fill((50, 150, 200))

            updates.append(S_BUTTONS)

        # print updates

        return updates


if __name__ == "__main__":
    pass



""" For storing the units information.  All in one file for ease of tweaking.
"""

from .const import *
from . import items


from . import cyclic_list


upgrade_amounts = {}
u = upgrade_amounts
cl = cyclic_list.cyclic_list

u["UpEngine Efficiency"] = cl(
    [1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0, 4.0, 4.0, 4.0]
)
u["UpEngine Speed"] = cl(
    [1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0, 4.0, 4.0, 4.0]
)
u["UpCannon Balls"] = cl(
    [1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0, 4.0, 4.0, 4.0]
)
u["UpArmour"] = cl(
    [1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0, 4.0, 4.0, 4.0]
)
u["UpSteam Tank"] = cl(
    [1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0, 4.0, 4.0, 4.0]
)
u["UpWater Tank"] = cl(
    [1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0, 4.0, 4.0, 4.0]
)
u["UpCoal Tank"] = cl(
    [1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0, 4.0, 4.0, 4.0]
)
u["UpCannon Power"] = cl(
    [1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 3.0, 3.0, 4.0, 4.0, 4.0, 4.0, 4.0]
)


upgrade_words = [
    "UpEngine Efficiency",
    "UpEngine Speed",
    "UpCannon Balls",
    "UpCannon Power",
    "UpArmour",
    "UpSteam Tank",
    "UpWater Tank",
    "UpCoal Tank",
]


class BaseUnit:
    def __init__(self, *args, **kwargs):
        self.stats = {
            "Armour": 0.0,
            "Speed": 20.0,
            "Weight": 1.0,
            "Health": 5.0,
            "Water": 10.0,
            "Coal": 10.0,
            "Steam": 0.0,
            "CannonPressure": 0.0,
            "MaxCannonPressure": 16.0,
            "EngineEfficiency": 1,
            "MaxHealth": 10.0,
            "MaxWater": 70.0,
            "MaxCoal": 70.0,
            "MaxSteam": 100.0,
            "Temperature": 0.0,
            "Damage": 1.0,
            "WeaponSteam": 1.0,
            "MoveSteam": 0.05,
            "GenerateSteamCoal": 0.05,
            "GenerateSteamWater": 0.05,
        }
        self.parts = {}

    def pickup_item(self, item):
        """this picks up the item passed in, and adds it to the robot."""

        if item.type == items.ITEM_COAL:
            if self.stats["Coal"] >= self.stats["MaxCoal"]:
                return 0
            else:
                self.stats["Coal"] += item.amount
                if self.stats["Coal"] >= self.stats["MaxCoal"]:
                    self.stats["Coal"] = self.stats["MaxCoal"]
                return 1

        if item.type == items.ITEM_WATER:
            if self.stats["Water"] >= self.stats["MaxWater"]:
                return 0
            else:
                self.stats["Water"] += item.amount
                if self.stats["Water"] >= self.stats["MaxWater"]:
                    self.stats["Water"] = self.stats["MaxWater"]

                return 1

        if item.type & items.ITEM_PART:
            self.parts[item.name] = item
            return 1

        return 0

    def hit(self, damage_amount):
        """ """
        # we can't do negative damage.
        x = damage_amount / (self.stats["Armour"] + 1.0)
        if x > 0:
            self.stats["Health"] -= x

        if self.stats["Health"] <= 0:
            return 1
        else:
            return 0

    def try_move(self):
        return self.try_use_steam(self.stats["MoveSteam"])

    def try_fire(self):
        # print "depricated"
        return self.try_use_steam(self.stats["WeaponSteam"])

    def prep_fire(self):
        self.stats["CannonPressure"] = 4

    def try_do_fire(self):
        if self.stats["CannonPressure"] > self.stats["MaxCannonPressure"]:
            self.stats["CannonPressure"] = self.stats["MaxCannonPressure"]

        if self.stats["CannonPressure"] > self.stats["Steam"]:
            self.stats["CannonPressure"] = self.stats["Steam"]

        if self.stats["CannonPressure"] < MIN_CANNON_PRESSURE:
            self.stats["CannonPressure"] = MIN_CANNON_PRESSURE

        if self.try_use_steam(self.stats["CannonPressure"]):
            pressure = self.stats["CannonPressure"]
            self.stats["CannonPressure"] = 0
            return pressure
        else:
            self.stats["CannonPressure"] = 0
            return MIN_CANNON_PRESSURE

    def try_use_steam(self, amount):
        """returns 1 if you can, else 0.  It uses up the steam regardless."""

        amount /= 2
        # print "try use steam."

        #
        if amount > self.stats["Steam"]:
            self.stats["Steam"] = 0.0
            return 0
        else:
            self.stats["Steam"] -= amount
            return 1

    def loop(self):
        self.generate_steam()

        if self.stats["CannonPressure"]:
            self.stats["CannonPressure"] += 1.2
            if self.stats["CannonPressure"] > self.stats["MaxCannonPressure"]:
                self.stats["CannonPressure"] = self.stats["MaxCannonPressure"]
            if self.stats["CannonPressure"] > self.stats["Steam"]:
                self.stats["CannonPressure"] = self.stats["Steam"]

    def generate_steam(self):
        """Returns amount of steam it generates from coal & water given."""

        coal = self.stats["GenerateSteamCoal"]
        water = self.stats["GenerateSteamWater"]

        # not allowed negatives...
        assert coal >= 0
        assert water >= 0

        # if no coal, or no water, then you can't have steam.
        if not coal or not water:
            print("no coal, or no water passed in for generation")
            steam = 0
            return 0
        else:
            # can only use what we have.
            if water > self.stats["Water"]:
                water = self.stats["Water"]

            if coal > self.stats["Coal"]:
                coal = self.stats["Coal"]

            self.stats["Water"] -= water
            self.stats["Coal"] -= coal

            if coal == 0 or water == 0:
                steam = 0
            else:
                # calculate how much steam we use.
                # ... this is just the average of the coal and water.
                steam = coal + water
                if steam > 0:
                    steam = steam / 2
                else:
                    pass
                    # print "no coal, or no water left"

        steam = steam * self.stats["EngineEfficiency"]

        total_steam = self.stats["Steam"] + steam
        if total_steam > self.stats["MaxSteam"]:
            # print "too much steam"
            self.stats["Steam"] = self.stats["MaxSteam"]
        else:
            self.stats["Steam"] += steam

        return steam

    def upgrade_part(self, part_up, amount):
        """call this to upgrade one part.
        part_up -
                  UpEngine Efficiency
                  UpEngine Speed
                  UpCannon Balls
                  UpArmour
                  UpSteam Tank
                  UpWater Tank
                  UpCoal Tank
                  UpCannon Power
        """

        print(("upgrading :%s" % part_up))

        if part_up == "UpEngine Efficiency":
            self.stats["EngineEfficiency"] += amount
        if part_up == "UpEngine Speed":
            self.stats["Speed"] += amount
        if part_up == "UpCannon Balls":
            self.stats["Damage"] += amount
        if part_up == "UpArmour":
            self.stats["Armour"] += amount
        if part_up == "UpSteam Tank":
            self.stats["MaxSteam"] += amount
        if part_up == "UpWater Tank":
            self.stats["MaxWater"] += amount
        if part_up == "UpCoal Tank":
            self.stats["MaxCoal"] += amount
        if part_up == "UpCannon Power":
            self.stats["MaxCannonPressure"] += amount


class Castle(BaseUnit):
    def __init__(self, *args, **kwargs):
        BaseUnit.__init__(self, *args, **kwargs)

        self.name = "castle"

        self.stats = {
            "Armour": 0.0,
            "Speed": 10.0,
            "Weight": 1.0,
            "Health": 10.0,
            "Water": 65.0,
            "Coal": 65.0,
            "Steam": 100.0,
            "EngineEfficiency": 3.0,
            "CannonPressure": 0.0,
            "MaxCannonPressure": 16.0,
            "MaxHealth": 10.0,
            "MaxWater": 70.0,
            "MaxCoal": 70.0,
            "MaxSteam": 100.0,
            "Temperature": 0.0,
            "Damage": 1.0,
            "WeaponSteam": 1.0,
            "MoveSteam": 0.05,
            "GenerateSteamCoal": 0.05,
            "GenerateSteamWater": 0.05,
            "UpEngine Speed": 0,
            "UpEngine Efficiency": 0,
            "UpCannon Balls": 0,
            "UpCannon Power": 0,
            "UpArmour": 0,
            "UpSteam Tank": 0,
            "UpWater Tank": 0,
            "UpCoal Tank": 0,
            "upgrade_amounts": upgrade_amounts,
        }


class Robot(BaseUnit):
    def __init__(self, *args, **kwargs):
        BaseUnit.__init__(self, *args, **kwargs)
        self.name = "baddie_robot"

        self.stats = {
            "Armour": 0.0,
            "Speed": 20.0,
            "Weight": 1.0,
            "Health": 5.0,
            "Water": 10.0,
            "Coal": 10.0,
            "EngineEfficiency": 1,
            "Steam": 100.0,
            "CannonPressure": 0.0,
            "MaxCannonPressure": 16.0,
            "MaxHealth": 70.0,
            "MaxWater": 70.0,
            "MaxCoal": 70.0,
            "MaxSteam": 70.0,
            "Temperature": 0.0,
            "Damage": 1.0,
            "WeaponSteam": 1.0,
            "MoveSteam": 0.05,
            "GenerateSteamCoal": 0.05,
            "GenerateSteamWater": 0.05,
        }


class CannonTower(BaseUnit):
    def __init__(self, *args, **kwargs):
        BaseUnit.__init__(self, *args, **kwargs)

        self.name = "cannon_tower"

        self.stats = {
            "Armour": 0.0,
            "Speed": 0.0,
            "Weight": 1.0,
            "Health": 3.0,
            "Water": 10.0,
            "Coal": 10.0,
            "Steam": 100.0,
            "MaxHealth": 70.0,
            "EngineEfficiency": 1,
            "MaxWater": 70.0,
            "MaxCoal": 70.0,
            "MaxSteam": 70.0,
            "CannonPressure": 0.0,
            "MaxCannonPressure": 16.0,
            "Temperature": 0.0,
            "Damage": 1.0,
            "WeaponSteam": 1.0,
            "MoveSteam": 0.05,
            "GenerateSteamCoal": 0.05,
            "GenerateSteamWater": 0.05,
        }


class CoalTruck(BaseUnit):
    def __init__(self, *args, **kwargs):
        BaseUnit.__init__(self, *args, **kwargs)
        self.name = "coal_truck"

        self.stats = {
            "Armour": 0.0,
            "Speed": 20.0,
            "Weight": 1.0,
            "Health": 5.0,
            "Water": 10.0,
            "Coal": 10.0,
            "Steam": 100.0,
            "EngineEfficiency": 1,
            "MaxHealth": 70.0,
            "MaxWater": 70.0,
            "MaxCoal": 70.0,
            "MaxSteam": 70.0,
            "CannonPressure": 0.0,
            "MaxCannonPressure": 16.0,
            "Temperature": 0.0,
            "Damage": 1.0,
            "WeaponSteam": 1.0,
            "MoveSteam": 0.05,
            "GenerateSteamCoal": 0.05,
            "GenerateSteamWater": 0.05,
        }


class Factory(BaseUnit):
    def __init__(self, *args, **kwargs):
        BaseUnit.__init__(self, *args, **kwargs)
        self.name = "factory"

        self.stats = {
            "Armour": 0.0,
            "Speed": 20.0,
            "Weight": 1.0,
            "Health": 5.0,
            "Water": 10.0,
            "EngineEfficiency": 1,
            "Coal": 10.0,
            "Steam": 100.0,
            "MaxHealth": 70.0,
            "MaxWater": 70.0,
            "MaxCoal": 70.0,
            "MaxSteam": 70.0,
            "CannonPressure": 0.0,
            "MaxCannonPressure": 16.0,
            "Temperature": 0.0,
            "Damage": 1.0,
            "WeaponSteam": 1.0,
            "MoveSteam": 0.05,
            "GenerateSteamCoal": 0.05,
            "GenerateSteamWater": 0.05,
        }



import time
import pygame
from pygame.locals import *
from .const import *

import os
from .pgu import engine

# tempo = 125
# speed = 6

data = [
    (0, "you", 0, 64),
    (8, "are", 90, 64),
    (16, "Zanthor", 128, 144),
    (28, "and", 160, 64),
    (32, "you", 190, 64),
    (48, "are", 210, 64),
    (64 + 0, "MAD", 255, 320),
]

#     (0,'you',0,32),
#     (8,'are',90,32),
#     (16,'Zanthor',128,72),
#     (28,'and',160,32),
#     (32,'you',190,32),
#     (48,'are',210,32),
#     (64+0,'MAD',255,160),

# uncomment this to test it without a mixer.
# pygame.mixer = None


class Intro(engine.State):
    def init(self):
        # rev up music...
        if pygame.mixer:
            pygame.mixer.music.load(data_dir("intro", "intro1.ogg"))
            # pygame.time.wait(2500) #let fullscreen kick in,
            pygame.mixer.music.play()
        else:
            self.cur_time = time.time()
            self.elapsed_time = 0.0

        im_path = data_dir("intro", "introbg.png")
        self.bkgr = pygame.image.load(im_path).convert()
        screen_size = pygame.display.get_surface().get_size()
        self.bkgr = pygame.transform.scale(self.bkgr, screen_size)

        self.fonts = {}

        for frame, text, alpha, size in data:
            self.fonts[size] = pygame.font.Font(data_dir("intro", "WALSHES.TTF"), size)

        self.cur = None
        self.frame = 0

        pygame.mouse.set_visible(False)

    def paint(self, screen):
        return
        # screen.fill((0,0,0))
        # pygame.display.flip()

    def loop(self):
        if pygame.mixer:
            if not pygame.mixer.music.get_busy():
                from . import title

                return title.Title(self.game)
        else:
            if self.elapsed_time > 15.3462460041:
                from . import title

                return title.Title(self.game)

    def update(self, screen):
        if pygame.mixer:
            t = pygame.mixer.music.get_pos()
        else:
            last_time = self.cur_time
            self.cur_time = time.time()
            self.elapsed_time += self.cur_time - last_time
            t = int(self.elapsed_time * 1000)

        f = t * 125 * 4 / (60 * 1000)

        for d in data:
            if f >= d[0]:
                cur = d

        # print f,'render',cur
        if cur != self.cur:
            self.cur, self.frame = cur, 0

        frame, text, alpha, size = cur

        if self.frame > 1 and text != "MAD":
            return

        fnt = self.fonts[size]

        if self.frame == 0:
            screen.fill((255, 255, 255))
            img = fnt.render(text, 1, (0, 0, 0))
            screen.blit(img, ((SW - img.get_width()) / 2, (SH - img.get_height()) / 2))

        if self.frame > 0:
            v = 0
            if alpha == 255:
                v = 255
            screen.fill((v, 0, 0))
            bg = self.bkgr
            if text == "MAD":
                alpha = max(0, alpha - self.frame * 4)
            bg.set_alpha(alpha)
            screen.blit(bg, (0, 0))
            img = fnt.render(text, 1, (0, 0, 0))
            for dx, dy in [(-2, 0), (2, 0), (0, 2), (0, -2)]:
                screen.blit(
                    img,
                    (dx + (SW - img.get_width()) / 2, dy + (SH - img.get_height()) / 2),
                )

            img = fnt.render(text, 1, (255, 255, 255))
            screen.blit(img, ((SW - img.get_width()) / 2, (SH - img.get_height()) / 2))

        pygame.display.flip()

        self.frame += 1

    def event(self, e):
        # if e.type in [KEYDOWN, MOUSEBUTTONDOWN, JOYBUTTONDOWN]:
        if e.type in [KEYDOWN, JOYBUTTONDOWN]:
            if pygame.mixer:
                pygame.mixer.music.stop()
            from . import title

            return title.Title(self.game)
            # return level.Level(self.game,0)



"""
"""
import pygame

from .const import *
from . import widget, surface


class Container(widget.Widget):
    """The base container widget, can be used as a template as well as stand alone.

    <pre>Container()</pre>
    """

    def __init__(self, **params):
        widget.Widget.__init__(self, **params)
        self.myfocus = None
        self.mywindow = None
        self.myhover = None
        # self.background = 0
        self.widgets = []
        self.windows = []
        self.toupdate = {}
        self.topaint = {}

    def update(self, s):
        updates = []

        if self.myfocus:
            self.toupdate[self.myfocus] = self.myfocus

        for w in self.topaint:
            if w is self.mywindow:
                continue
            else:
                w.paint(surface.subsurface(s, w.rect))
                updates.append(pygame.rect.Rect(w.rect))

        for w in self.toupdate:
            if w is self.mywindow:
                continue
            else:
                us = w.update(surface.subsurface(s, w.rect))
            if us:
                for u in us:
                    updates.append(
                        pygame.rect.Rect(u.x + w.rect.x, u.y + w.rect.y, u.w, u.h)
                    )

        for w in self.topaint:
            if w is self.mywindow:
                w.paint(self.top_surface(s, w))
                updates.append(pygame.rect.Rect(w.rect))
            else:
                continue

        for w in self.toupdate:
            if w is self.mywindow:
                us = w.update(self.top_surface(s, w))
            else:
                continue
            if us:
                for u in us:
                    updates.append(
                        pygame.rect.Rect(u.x + w.rect.x, u.y + w.rect.y, u.w, u.h)
                    )

        self.topaint = {}
        self.toupdate = {}

        return updates

    def repaint(self, w=None):
        if not w:
            return widget.Widget.repaint(self)
        self.topaint[w] = w
        self.reupdate()

    def reupdate(self, w=None):
        if not w:
            return widget.Widget.reupdate(self)
        self.toupdate[w] = w
        self.reupdate()

    def paint(self, s):
        self.toupdate = {}
        self.topaint = {}

        for w in self.widgets:
            ok = False
            try:
                sub = surface.subsurface(s, w.rect)
                ok = True
            except:
                print(
                    "container.paint(): %s not in %s"
                    % (w.__class__.__name__, self.__class__.__name__)
                )
                print(s.get_width(), s.get_height(), w.rect)
                ok = False
            if ok:
                w.paint(sub)

        for w in self.windows:
            w.paint(self.top_surface(s, w))

    def top_surface(self, s, w):
        x, y = s.get_abs_offset()
        s = s.get_abs_parent()
        return surface.subsurface(s, (x + w.rect.x, y + w.rect.y, w.rect.w, w.rect.h))

    def event(self, e):
        if self.mywindow and e.type == MOUSEBUTTONDOWN:
            w = self.mywindow
            if self.myfocus is w:
                if not w.rect.collidepoint(e.pos):
                    self.blur(w)
            if not self.myfocus:
                if w.rect.collidepoint(e.pos):
                    self.focus(w)

        if not self.mywindow:
            #### by Gal Koren
            ##
            ## if e.type == FOCUS:
            if e.type == FOCUS and not self.myfocus:
                # self.first()
                pass
            elif e.type == EXIT:
                if self.myhover:
                    self.exit(self.myhover)
            elif e.type == BLUR:
                if self.myfocus:
                    self.blur(self.myfocus)
            elif e.type == MOUSEBUTTONDOWN:
                h = None
                for w in self.widgets:
                    if w.rect.collidepoint(e.pos):
                        h = w
                        if self.myfocus is not w:
                            self.focus(w)
                if not h and self.myfocus:
                    self.blur(self.myfocus)
            elif e.type == MOUSEMOTION:
                if 1 in e.buttons:
                    if self.myfocus:
                        ws = [self.myfocus]
                    else:
                        ws = []
                else:
                    ws = self.widgets

                h = None
                for w in ws:
                    if w.rect.collidepoint(e.pos):
                        h = w
                        if self.myhover is not w:
                            self.enter(w)
                if not h and self.myhover:
                    self.exit(self.myhover)
                w = self.myhover

                if w and w is not self.myfocus:
                    sub = pygame.event.Event(
                        e.type,
                        {
                            "buttons": e.buttons,
                            "pos": (e.pos[0] - w.rect.x, e.pos[1] - w.rect.y),
                            "rel": e.rel,
                        },
                    )
                    w._event(sub)

        w = self.myfocus
        if w:
            sub = e

            if e.type == MOUSEBUTTONUP or e.type == MOUSEBUTTONDOWN:
                sub = pygame.event.Event(
                    e.type,
                    {
                        "button": e.button,
                        "pos": (e.pos[0] - w.rect.x, e.pos[1] - w.rect.y),
                    },
                )
                w._event(sub)
            elif e.type == CLICK and self.myhover is w:
                sub = pygame.event.Event(
                    e.type,
                    {
                        "button": e.button,
                        "pos": (e.pos[0] - w.rect.x, e.pos[1] - w.rect.y),
                    },
                )
                w._event(sub)
            elif e.type == CLICK:  # a dead click
                pass
            elif e.type == MOUSEMOTION:
                sub = pygame.event.Event(
                    e.type,
                    {
                        "buttons": e.buttons,
                        "pos": (e.pos[0] - w.rect.x, e.pos[1] - w.rect.y),
                        "rel": e.rel,
                    },
                )
                w._event(sub)
            else:
                w._event(sub)

    def remove(self, w):
        """Remove a widget from the container.

        <pre>Container.remove(w)</pre>
        """
        self.blur(w)
        self.widgets.remove(w)
        # self.repaint()
        self.chsize()

    def add(self, w, x, y):
        """Add a widget to the container.

        <pre>Container.add(w,x,y)</pre>

        <dl>
        <dt>x, y<dd>position of the widget
        </dl>
        """
        w.style.x = x
        w.style.y = y
        w.container = self
        # NOTE: this might fix it, sort of...
        # but the thing is, we don't really want to resize
        # something if it is going to get resized again later
        # for no reason...
        # w.rect.x,w.rect.y = w.style.x,w.style.y
        # w.rect.w, w.rect.h = w.resize()
        self.widgets.append(w)
        self.chsize()

    def open(self, w=None, x=None, y=None):
        from .app import App  # HACK: I import it here to prevent circular importing

        if not w:
            if (
                not hasattr(self, "container") or not self.container
            ) and self is not App.app:
                self.container = App.app
            # print 'top level open'
            return widget.Widget.open(self)

        if self.container:
            if x != None:
                return self.container.open(w, self.rect.x + x, self.rect.y + y)
            return self.container.open(w)

        w.container = self

        if w.rect.w == 0 or w.rect.h == 0:  # this might be okay, not sure if needed.
            w.rect.w, w.rect.h = w.resize()

        if x == None or y == None:  # auto center the window
            # w.style.x,w.style.y = 0,0
            w.rect.x = (self.rect.w - w.rect.w) / 2
            w.rect.y = (self.rect.h - w.rect.h) / 2
            # w.resize()
            # w._resize(self.rect.w,self.rect.h)
        else:  # show it where we want it
            w.rect.x = x
            w.rect.y = y
            # w._resize()

        self.windows.append(w)
        self.mywindow = w
        self.focus(w)
        self.repaint(w)
        w.send(OPEN)

    def close(self, w=None):
        if not w:
            return widget.Widget.close(self)

        if self.container:  # make sure we're in the App
            return self.container.close(w)

        if self.myfocus is w:
            self.blur(w)

        if w not in self.windows:
            return  # no need to remove it twice! happens.

        self.windows.remove(w)

        self.mywindow = None
        if self.windows:
            self.mywindow = self.windows[-1]
            self.focus(self.mywindow)

        if not self.mywindow:
            self.myfocus = self.widget  # HACK: should be done fancier, i think..
            if not self.myhover:
                self.enter(self.widget)

        self.repaintall()
        w.send(CLOSE)

    def focus(self, w=None):
        widget.Widget.focus(self)  ### by Gal koren
        #        if not w:
        #            return widget.Widget.focus(self)
        if not w:
            return
        if self.myfocus:
            self.blur(self.myfocus)
        if self.myhover is not w:
            self.enter(w)
        self.myfocus = w
        w._event(pygame.event.Event(FOCUS))

    def blur(self, w=None):
        if not w:
            return widget.Widget.blur(self)
        if self.myfocus is w:
            if self.myhover is w:
                self.exit(w)
            self.myfocus = None
            w._event(pygame.event.Event(BLUR))

    def enter(self, w):
        if self.myhover:
            self.exit(self.myhover)
        self.myhover = w
        w._event(pygame.event.Event(ENTER))

    def exit(self, w):
        if self.myhover and self.myhover is w:
            self.myhover = None
            w._event(pygame.event.Event(EXIT))

    #     def first(self):
    #         for w in self.widgets:
    #             if w.focusable:
    #                 self.focus(w)
    #                 return
    #         if self.container: self.container.next(self)

    #     def next(self,w):
    #         if w not in self.widgets: return #HACK: maybe.  this happens in windows for some reason...
    #
    #         for w in self.widgets[self.widgets.index(w)+1:]:
    #             if w.focusable:
    #                 self.focus(w)
    #                 return
    #         if self.container: return self.container.next(self)

    def _next(self, orig=None):
        start = 0
        if orig in self.widgets:
            start = self.widgets.index(orig) + 1
        for w in self.widgets[start:]:
            if not w.disabled and w.focusable:
                if isinstance(w, Container):
                    if w._next():
                        return True
                else:
                    self.focus(w)
                    return True
        return False

    def next(self, w):
        if w not in self.widgets:
            return  # HACK: maybe.  this happens in windows for some reason...

        if self._next(w):
            return True
        if self.container:
            return self.container.next(self)

    def previous(self, w):
        print("Container.previous: n/a")

    def resize(self, width=None, height=None):
        # r = self.rect
        # r.w,r.h = 0,0
        ww, hh = 0, 0
        if self.style.width:
            ww = self.style.width
        if self.style.height:
            hh = self.style.height

        for w in self.widgets:
            # w.rect.w,w.rect.h = 0,0
            w.rect.x, w.rect.y = w.style.x, w.style.y
            w.rect.w, w.rect.h = w.resize()
            # w._resize()

            ww = max(ww, w.rect.right)
            hh = max(hh, w.rect.bottom)
        return ww, hh



import pygame
from pygame.locals import *
import random

from .pgu import timer

rr = random.randrange


class Part:
    def __init__(self, x, y):
        x, y = float(x), float(y)
        self.x, self.y = x, y
        self._x, self._y = x, y
        self.min = 4.0
        self.max = 16.0
        self.radius = 16.0
        self.center = 2.0
        self.search = 64.0
        self._flock_frame = 0
        self.near = []


class Flock:
    def __init__(self, rect, spacing):
        self.rect, self.spacing = rect, spacing
        self.grid = [
            [[] for tx in range(rect.left, rect.right, spacing)]
            for ty in range(rect.top, rect.bottom, spacing)
        ]
        self.parts = []

    def append(self, p):
        rect, spacing, grid, parts = self.rect, self.spacing, self.grid, self.parts
        self.parts.append(p)
        tx, ty = p._tx, p._ty = int((p.x - rect.left) / spacing), int(
            (p.y - rect.top) / spacing
        )
        p.vx, p.vy = 0, 0
        grid[ty][tx].append(p)

    def remove(self, p):
        rect, spacing, grid, parts = self.rect, self.spacing, self.grid, self.parts
        tx, ty = p._tx, p._ty
        grid[ty][tx].remove(p)
        parts.remove(p)

    def __len__(self):
        return self.parts.__len__()

    def __iter__(self):
        return self.parts.__iter__()

    def loop(self):
        rect, spacing, grid, parts = self.rect, self.spacing, self.grid, self.parts
        gw, gh = len(grid[0]), len(grid)

        # update grid
        for p in parts:
            tx, ty = p._tx, p._ty
            grid[ty][tx].remove(p)
            tx, ty = p._tx, p._ty = int((p.x - rect.left) / spacing), int(
                (p.y - rect.top) / spacing
            )
            grid[ty][tx].append(p)

        # calculate velocities
        for p in parts:
            p.vx, p.vy = p.x - p._x, p.y - p._y
            p._x, p._y = p.x, p.y

        # calculate neighbors
        for p in parts:
            tx, ty = p._tx, p._ty

            p.near = []
            for dy in (-1, 0, 1):
                for dx in (-1, 0, 1):
                    xx, yy = tx + dx, ty + dy
                    if xx < 0 or yy < 0 or xx >= gw or yy >= gh:
                        continue
                    p.near.extend(grid[yy][xx])
            p.near.remove(p)

            p.ax, p.ay = p.x, p.y
            p.avx, p.avy = p.vx, p.vy

            for _p in p.near:
                p.ax += _p.x
                p.ay += _p.y
                p.avx += _p.vx
                p.avy += _p.vy

            total = len(p.near) + 1
            p.ax /= total
            p.ay /= total
            p.avx /= total
            p.avy /= total

        # add average velocity to item
        for p in parts:
            p.x, p.y = p.x + p.avx, p.y + p.avy

        # move towards the center
        for p in parts:
            step = p.center
            ax, ay = p.ax, p.ay
            dx, dy = ax - p.x, ay - p.y
            dist = (dx * dx + dy * dy) ** 0.5
            if dist > step:
                p.x, p.y = p.x + step * dx / dist, p.y + step * dy / dist

        # check min velocity
        for p in parts:
            step = p.min
            dx, dy = p.x - p._x, p.y - p._y
            dist = (dx * dx + dy * dy) ** 0.5
            while dist < 0.5:
                dx, dy = random.randrange(-1, 2), random.randrange(-1, 2)
                dist = (dx * dx + dy * dy) ** 0.5
            if dist < step:
                p.x, p.y = p._x + dx * step / dist, p._y + dy * step / dist

        # keep away from neighbords
        for p in parts:
            for _p in p.near:
                dx, dy = _p.x - p.x, _p.y - p.y
                dist = (dx * dx + dy * dy) ** 0.5
                r = p.radius + _p.radius
                if dist < r:
                    if dist == 0:
                        dist, dx = 1, 1
                    inc = r / 2.0
                    mx, my = (p.x + _p.x) / 2, (p.y + _p.y) / 2
                    p.x, p.y = mx - (inc * dx / dist), my - (inc * dy / dist)
                    _p.x, _p.y = mx + (inc * dx / dist), my + (inc * dy / dist)

        # check max velocity
        for p in parts:
            step = p.max
            dx, dy = p.x - p._x, p.y - p._y
            dist = (dx * dx + dy * dy) ** 0.5
            if dist > step:
                p.x, p.y = p._x + dx * step / dist, p._y + dy * step / dist


if __name__ == "__main__":
    try:
        # 0/0 #HACK: so i can CTRL-c out
        import psyco

        psyco.profile()
        print("psyco installed")
    except:
        print("psyco not installed")

    SW, SH = 640, 480
    screen = pygame.display.set_mode((SW, SH))

    num = 256
    rad = 12
    spacing = rad * 3 / 2
    border = spacing * 2
    myflock = Flock(
        pygame.Rect(-border, -border, SW + border * 2, SH + border * 2), spacing
    )
    for n in range(0, num):
        p = Part(rr(0, SW), rr(0, SH))
        p.radius = rad
        myflock.append(p)

    t = timer.Speedometer()
    pygame.time.get_ticks()

    us = [(0, 0, SW, SH)]

    _quit = False
    frames = 0
    while not _quit:
        for e in pygame.event.get():
            if e.type == QUIT:
                _quit = True
            if e.type == KEYDOWN and e.key == K_ESCAPE:
                _quit = True

        for r in us:
            screen.fill((0, 0, 0), r)
        _us, us = us, []
        us = []

        p = myflock.parts[0]  # HACK!!
        p.x, p.y = pygame.mouse.get_pos()
        p.radius = 64
        x, y, r = int(p.x), int(p.y), int(p.radius)
        pygame.draw.circle(screen, (0, 0, 128), (x, y), r)
        us.append((x - r, y - r, r * 2, r * 2))

        myflock.loop()

        for p in myflock:  # HACK!!
            p.x = max(0, min(SW, p.x))
            p.y = max(0, min(SH, p.y))
            rect = int(p.x), int(p.y), 4, 4
            screen.fill((255, 255, 255), rect)
            us.append(rect)
        pygame.display.update(_us)
        pygame.display.update(us)

        r = t.tick()
        if r:
            print(r)
        frames += 1



"""
"""
from .const import *
from . import table
from . import basic, button


class Select(table.Table):
    """A select input.

    <pre>Select(value=None)</pre>

    <dl>
    <dt>value<dd>initial value
    </dl>

    <strong>Example</strong>
    <code>
    w = Select(value="goats")
    w.add("Cats","cats")
    w.add("Goats","goats")
    w.add("Dogs","Dogs")

    w.value = 'dogs' #changes the value from goats to dogs
    </code>

    """

    def __init__(self, value=None, **params):
        params.setdefault("cls", "select")
        table.Table.__init__(self, **params)

        self.top_selected = button.Button(cls=self.cls + ".selected")
        table.Table.add(self, self.top_selected)  # ,hexpand=1,vexpand=1)#,0,0)
        self.top_selected.value = basic.Label(" ", cls=self.cls + ".option.label")

        self.top_arrow = button.Button(
            basic.Image(self.style.arrow), cls=self.cls + ".arrow"
        )
        table.Table.add(self, self.top_arrow)  # ,hexpand=1,vexpand=1) #,1,0)

        self.options = table.Table()  # style={'border':3})

        self.options.tr()
        self.spacer_top = basic.Spacer(0, 0)
        self.options.add(self.spacer_top)

        self.options.tr()
        self._options = table.Table(cls=self.cls + ".options")
        self.options.add(self._options)

        self.options.tr()
        self.spacer_bottom = basic.Spacer(0, 0)
        self.options.add(self.spacer_bottom)

        self.options.connect(BLUR, self.options.close, None)
        self.spacer_top.connect(CLICK, self.options.close, None)
        self.spacer_bottom.connect(CLICK, self.options.close, None)

        self.values = []
        self.value = value

    def resize(self, width=None, height=None):
        max_w, max_h = 0, 0
        for w in self._options.widgets:
            w.rect.w, w.rect.h = w.resize()
            max_w, max_h = max(max_w, w.rect.w), max(max_h, w.rect.h)

        # xt,xr,xb,xl = self.top_selected.getspacing()
        self.top_selected.style.width = max_w  # + xl + xr
        self.top_selected.style.height = max_h  # + xt + xb

        self.top_arrow.connect(CLICK, self._open, None)
        self.top_selected.connect(CLICK, self._open, None)

        w, h = table.Table.resize(self, width, height)

        self.spacer_top.style.width, self.spacer_top.style.height = w, h
        self.spacer_bottom.style.width, self.spacer_bottom.style.height = w, h
        self._options.style.width = w - 2  # HACK, this depends on the theme, really...
        # HACK: sort of, but not a big one..
        self._options.resize()

        return w, h

    def _open(self, value):
        sh = self.rect.h  # spacer height
        opts = self.options

        self.spacer_top.style.height = 0
        self.spacer_bottom.style.height = 0
        opts.rect.w, opts.rect.h = opts.resize()
        h = opts.rect.h

        y = self.rect.y
        c = self.container
        while hasattr(c, "container"):
            y += c.rect.y
            if c.container == None:
                break
            c = c.container

        if y + sh + h <= c.rect.h:  # down
            self.spacer_top.style.height = sh
            dy = self.rect.y
        else:  # up
            self.spacer_bottom.style.height = sh
            dy = self.rect.y - h

        opts.rect.w, opts.rect.h = opts.resize()

        self.container.open(opts, self.rect.x, dy)

    def _setvalue(self, value):
        self.value = value._value
        if hasattr(self, "container"):
            # self.chsize()
            # HACK: improper use of resize()
            # self.resize() #to recenter the new value, etc.
            pass
        #    #self._resize()

        self.options.close()
        self.repaint()  # this will happen anyways

    def __setattr__(self, k, v):
        mywidget = None
        if k == "value":
            for w in self.values:
                if w._value == v:
                    mywidget = w
        _v = self.__dict__.get(k, NOATTR)
        self.__dict__[k] = v
        if k == "value" and _v != NOATTR and _v != v:
            self.send(CHANGE)
            self.repaint()
        if k == "value":
            if not mywidget:
                mywidget = basic.Label(" ", cls=self.cls + ".option.label")
            self.top_selected.value = mywidget

    def add(self, w, value=None):
        """Add a widget, value item to the Select.

        <pre>Select.add(widget,value=None)</pre>

        <dl>
        <dt>widget<dd>Widget or string to represent the item
        <dt>value<dd>value for this item
        </dl>

        <strong>Example</strong>
        <code>
        w = Select()
        w.add("Goat") #adds a Label
        w.add("Goat","goat") #adds a Label with the value goat
        w.add(gui.Label("Cuzco"),"goat") #adds a Label with value goat
        </code>
        """

        if type(w) == str:
            w = basic.Label(w, cls=self.cls + ".option.label")

        w.style.align = -1
        b = button.Button(w, cls=self.cls + ".option")
        b.connect(CLICK, self._setvalue, w)
        # b = w
        # w.cls = self.cls+".option"
        # w.cls = self.cls+".option"

        self._options.tr()
        self._options.add(b)  # ,align=-1)
        # self._options.td(b, align=-1, cls=self.cls+".option")
        # self._options.td(_List_Item(w,value=value),align=-1)

        if value != None:
            w._value = value
        else:
            w._value = w
        if self.value == w._value:
            self.top_selected.value = w
        self.values.append(w)



"""
"""


class Style:
    """The class used by widget for the widget.style

    <p>This object is used mainly as a dictionary, accessed via <tt>widget.style.attr</tt>, as opposed to
    <tt>widget.style['attr']</tt>.  It automatically grabs information from the theme via <tt>value = theme.get(widget.cls,widget.pcls,attr)</tt>.</p>

    """

    def __init__(self, o, dict):
        self.obj = o
        self._cache = 0
        for k, v in list(dict.items()):
            self.__dict__[k] = v

    def __getattr__(self, k):
        from . import app

        v = app.App.app.theme.get(self.obj.cls, self.obj.pcls, k)
        if self._cache:
            self.__dict__[k] = v
        return v

    def cache(self, v):
        """Enable the cache.

        <p>This should be used when changes of the pcls are unlikely to impact the data, or when there is a lot of repetitive access to the same attributes expected.</p>

        <pre>Style.cache(v)</pre>

        <dl>
        <dt>v<dd>True if the cache should be enabled.
        </dl>
        """
        if v:
            self._cache += 1
        else:
            self._cache -= 1

    def __setattr__(self, k, v):
        self.__dict__[k] = v



"""Some handy font-like objects.

<p>please note that this file is alpha, and is subject to modification in
future versions of pgu!</p>
"""

print("pgu.fonts", "This module is alpha, and is subject to change.")

import pygame
from pygame.locals import *


class TileFont:
    """Creates an instance of the TileFont class.  Interface compatible with pygame.Font

    <p>TileFonts are fonts that are stored in a tiled image.  Where the image opaque, it assumed that the font is visible.  Font color is changed automatically, so it does not work with
    fonts with stylized coloring.</p>

    <pre>TileFont(fname,size,hints,scale=None,sensitive=False)</pre>

    <dl>
    <dt>size <dd>the dimensions of the characters
    <dt>hints <dd>a string of hints "abcdefg..."
    <dt>scale <dd>size to scale font to
    <dt>sensitive <dd>case sensitivity
    </dl>
    """

    def __init__(self, fname, size, hints, scale=None, sensitive=False):
        self.image = pygame.image.load(fname)

        w, h = self.image.get_width(), self.image.get_height()
        tw, th = size
        if not scale:
            scale = size
        self._size = size
        self.scale = scale

        self.chars = {}
        x, y = 0, 0
        self.sensitive = sensitive
        if not self.sensitive:
            hints = hints.lower()
        for c in hints:
            if c not in ("\r", "\n", "\t"):
                img = self.image.subsurface(x, y, tw, th)
                self.chars[c] = img
                x += tw
                if x >= w:
                    x, y = 0, y + th

        self.colors = {}

    def size(self, text):
        tw, th = self.scale
        return len(text) * tw, th

    def render(self, text, antialias=0, color=(255, 255, 255), background=None):
        size = self.size(text)
        scale = self.scale
        tw, th = self._size
        if background == None:
            s = pygame.Surface(size).convert_alpha()
            s.fill((0, 0, 0, 0))
        else:
            s = pygame.Surface(size).convert()
            s.fill(background)

        if not self.sensitive:
            text = text.lower()

        if color not in self.colors:
            self.colors[color] = {}
        colored = self.colors[color]

        x, y = 0, 0
        for c in text:
            if c in self.chars:
                if c not in colored:
                    img = self.chars[c].convert_alpha()
                    for yy in range(0, th):
                        for xx in range(0, tw):
                            r, g, b, a = img.get_at((xx, yy))
                            if a > 128:
                                img.set_at((xx, yy), color)
                    colored[c] = img
                img = colored[c]
                if scale != (tw, th):
                    img = pygame.transform.scale(img, scale)
                s.blit(img, (x, y))
            x += scale[0]
        return s


class BorderFont:
    """a decorator for normal fonts, adds a border. Interface compatible with pygame.Font.

    <pre>BorderFont(font,size=1,color=(0,0,0))</pre>

    <dl>
    <dt>size <dd>width of border; defaults 0
    <dt>color <dd>color of border; default (0,0,0)
    </dl>
    """

    def __init__(self, font, size=1, color=(0, 0, 0)):
        self.font = font
        self._size = size
        self.color = color

    def size(self, text):
        w, h = self.font.size(text)
        s = self._size
        return w + s * 2, h + s * 2

    def render(self, text, antialias=0, color=(255, 255, 255), background=None):
        size = self.size(text)

        if background == None:
            s = pygame.Surface(size).convert_alpha()
            s.fill((0, 0, 0, 0))
        else:
            s = pygame.Surface(size).convert()
            s.fill(background)

        bg = self.font.render(text, antialias, self.color)
        fg = self.font.render(text, antialias, color)

        si = self._size
        dirs = [(-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 1), (1, -1), (1, 0), (1, 1)]
        for dx, dy in dirs:
            s.blit(bg, (si + dx * si, si + dy * si))
        s.blit(fg, (si, si))

        return s



"""
"""
import pygame

from . import style


class Widget:
    """Template object - base for all widgets.

    <pre>Widget(**params)</pre>

    <p>A number of optional params may be passed to the Widget initializer.</p>

    <dl>
    <dt>decorate<dd>defaults to True.  If true, will call <tt>theme.decorate(self)</tt> to allow the theme a chance to decorate the widget.
    <dt>style<dd>a dict of style parameters.
    <dt>x, y, width, height<dd>position and size parameters, passed along to style
    <dt>align, valign<dd>alignment parameters, passed along to style
    <dt>font, color, background<dd>other common parameters that are passed along to style
    <dt>cls<dd>class name as used by Theme
    <dt>name<dd>name of widget as used by Form.  If set, will call <tt>form.add(self,name)</tt> to add the widget to the most recently created Form.
    <dt>focusable<dd>True if this widget can receive focus.  Defaults to True.
    <dt>disabled<dd>True of this widget is disabled.  Defaults to False.
    <dt>value<dd>initial value
    </dl>

    <strong>Example - Creating your own Widget</strong>
    <p>This example shows which methods are template methods.</p>
    <code>
    class Draw(gui.Widget):
        def paint(self,s):
            #paint the pygame.Surface
            return

        def update(self,s):
            #update the pygame.Surface and return the update rects
            return [pygame.Rect(0,0,self.rect.w,self.rect.h)]

        def event(self,e):
            #handle the pygame.Event
            return

        def resize(self,width=None,height=None):
            #return the width and height of this widget
            return 256,256
    </code>
    """

    def __init__(self, **params):
        # object.Object.__init__(self)
        self.connects = {}
        params.setdefault("decorate", True)
        params.setdefault("style", {})
        params.setdefault("focusable", True)
        params.setdefault("disabled", False)

        self.focusable = params["focusable"]
        self.disabled = params["disabled"]

        self.rect = pygame.Rect(
            params.get("x", 0),
            params.get("y", 0),
            params.get("width", 0),
            params.get("height", 0),
        )

        s = params["style"]
        # some of this is a bit "theme-ish" but it is very handy, so these
        # things don't have to be put directly into the style.
        for att in (
            "align",
            "valign",
            "x",
            "y",
            "width",
            "height",
            "color",
            "font",
            "background",
        ):
            if att in params:
                s[att] = params[att]
        self.style = style.Style(self, s)

        self.cls = "default"
        if "cls" in params:
            self.cls = params["cls"]
        if "name" in params:
            from . import form

            self.name = params["name"]
            if hasattr(form.Form, "form") and form.Form.form != None:
                form.Form.form.add(self)
                self.form = form.Form.form
        if "value" in params:
            self.value = params["value"]
        self.pcls = ""

        if params["decorate"] != False:
            from . import app

            if not hasattr(app.App, "app"):
                print("gui.widget: creating an App")
                app.App.app = app.App()
            app.App.app.theme.decorate(self, params["decorate"])

    def focus(self):
        """Focus this Widget.

        <pre>Widget.focus()</pre>
        """
        if getattr(self, "container", None) != None:
            if self.container.myfocus != self:  ## by Gal Koren
                self.container.focus(self)

    def blur(self):
        """Blur this Widget.

        <pre>Widget.blur()</pre>
        """
        if getattr(self, "container", None) != None:
            self.container.blur(self)

    def open(self):
        """Open this Widget as a modal dialog.

        <pre>Widget.open()</pre>
        """
        if getattr(self, "container", None) != None:
            self.container.open(self)

    def close(self):
        """Close this Widget (if it is a modal dialog.)

        <pre>Widget.close()</pre>
        """
        if getattr(self, "container", None) != None:
            self.container.close(self)

    def resize(self, width=None, height=None):
        """Template method - return the size and width of this widget.

        <p>Responsible for also resizing all sub-widgets.</p>

        <pre>Widget.resize(width,height): return width,height</pre>

        <dl>
        <dt>width<dd>suggested width
        <dt>height<dd>suggested height
        </dl>

        <p>If not overridden, will return self.style.width, self.style.height</p>
        """
        return self.style.width, self.style.height

    def chsize(self):
        """Change the size of this widget.

        <p>Calling this method will cause a resize on all the widgets,
        including this one.</p>

        <pre>Widget.chsize()</pre>
        """
        if not hasattr(self, "container"):
            return
        from . import app

        if hasattr(app.App, "app"):
            if app.App.app._chsize:
                return
            app.App.app.chsize()
            return

        # if hasattr(app.App,'app'):
        #    w,h = self.rect.w,self.rect.h
        #    w2,h2 = self.resize()
        #    if w2 != w or h2 != h:
        #        app.App.app.chsize()
        #    else:
        #        self.repaint()

    def update(self, s):
        """Template method - update the surface

        <pre>Widget.update(s): return list of pygame.Rect(s)</pre>

        <dl>
        <dt>s<dd>pygame.Surface to update
        </dl>

        <p>return - a list of the updated areas as pygame.Rect(s).</p>
        """
        return

    def paint(self, s):
        """Template method - paint the surface

        <pre>Widget.paint(s)</pre>

        <dl>
        <dt>s<dd>pygame.Surface to paint
        </dl>
        """
        return

    def repaint(self):
        """Request a repaint of this Widget.

        <pre>Widget.repaint()</pre>
        """
        if getattr(self, "container", None) != None:
            self.container.repaint(self)

    def repaintall(self):
        """Request a repaint of all Widgets.

        <pre>Widget.repaintall()</pre>
        """
        if getattr(self, "container", None) != None:
            self.container.repaintall()

    def reupdate(self):
        """Request a reupdate of this Widget

        <pre>Widget.reupdate()</pre>
        """
        if getattr(self, "container", None) != None:
            self.container.reupdate(self)

    def __next__(self):
        """Pass focus to next Widget.

        <p>Widget order determined by the order they were added to their container.</p>

        <pre>Widget.next()</pre>
        """
        if getattr(self, "container", None) != None:
            self.container.next(self)

    def previous(self):
        """Pass focus to previous Widget.

        <p>Widget order determined by the order they were added to their container.</p>

        <pre>Widget.previous()</pre>
        """

        if getattr(self, "container", None) != None:
            self.container.previous(self)

    def get_abs_rect(self):
        """Get the absolute rect of this widget on the App screen

        <pre>Widget.get_abs_rect(): return pygame.Rect</pre>
        """
        x, y = self.rect[:2]
        c = getattr(self, "container", None)
        while c:
            x += c.rect[0]
            y += c.rect[1]
            c = getattr(c, "container", None)
        return pygame.Rect(x, y, self.rect.w, self.rect.h)

    def connect(self, code, fnc, *values):
        """Connect a event code to a callback function.

        <p>There may only be one callback per event code.</p>

        <pre>Object.connect(code,fnc,value)</pre>

        <dl>
        <dt>code<dd>event type [[gui-const]]
        <dt>fnc<dd>callback function
        <dt>*values<dd>values to pass to callback.  Please note that callbacks may also have "magicaly" parameters.  Such as:
            <dl>
            <dt>_event<dd>receive the event
            <dt>_code<dd>receive the event code
            <dt>_widget<dd>receive the sending widget
            </dl>
        </dl>

        <strong>Example</strong>
        <code>
        def onclick(value):
            print 'click',value

        w = Button("PGU!")
        w.connect(gui.CLICK,onclick,'PGU Button Clicked')
        </code>
        """

        self.connects[code] = {"fnc": fnc, "values": values}

    def send(self, code, event=None):
        """Send a code, event callback trigger.

        <pre>Object.send(code,event=None)</pre>

        <dl>
        <dt>code<dd>event code
        <dt>event<dd>event
        </dl>
        """
        if code in self.connects:
            con = self.connects[code]
            # con['fnc'](*con['values'])

            fnc = con["fnc"]
            values = list(con["values"])

            nargs = fnc.__code__.co_argcount
            names = list(fnc.__code__.co_varnames)[:nargs]
            if hasattr(fnc, "im_class"):
                names.pop(0)

            args = []
            magic = {"_event": event, "_code": code, "_widget": self}
            for name in names:
                if name in list(magic.keys()):
                    args.append(magic[name])
                elif len(values):
                    args.append(values.pop(0))
                else:
                    break
            args.extend(values)
            fnc(*args)

    def _event(self, e):
        if self.disabled:
            return
        self.send(e.type, e)
        self.event(e)
        return
        from . import app

        if hasattr(app.App, "app"):
            app.App.app.events.append((self, e))

    def event(self, e):
        """Template method - called when an event is passed to this object.

        <dl>
        <dt>e<dd>event
        </dl>
        """

        return



ITEM_COAL = 0x0001
ITEM_WATER = 0x0002
ITEM_PART = 0x0010
ITEM_CANNON = 0x0020
ITEM_ENGINE = 0x0030
ITEM_WATERTANK = 0x0040
ITEM_COALSTORAGEROOM = 0x0050
ITEM_ARMOUR = 0x0060
ITEM_RUBBLE = 0x0100
ITEM_ = 10
ITEM_ = 11
ITEM_ = 12
ITEM_ = 13
ITEM_ = 14
ITEM_ = 15
ITEM_ = 16


class Coal:
    type = ITEM_COAL

    def __init__(self, amount=1.0):
        self.amount = amount
        self.name = "Coal"
        self.pickup_string = "Pick up some Coal?"


class Water:
    type = ITEM_WATER

    def __init__(self, amount=1.0):
        self.amount = amount
        self.name = "Water"
        self.pickup_string = "Suck up some water?"


class Rubble:
    type = ITEM_RUBBLE

    def __init__(self):
        self.name = "Rubble"
        self.pickup_string = "Pick up some rubble?"


class Cannon:
    type = ITEM_CANNON

    def __init__(self):
        self.name = "cannon"
        self.pickup_string = "Pick up a Cannon?"


class Part:
    type = ITEM_PART

    def __init__(self, name, pickup_string):
        self.name = name
        self.pickup_string = pickup_string


class Engine:
    type = ITEM_ENGINE

    def __init__(self):
        self.name = "engine"
        self.pickup_string = "Pick up an engine?"


class WaterTank:
    type = ITEM_WATERTANK

    def __init__(self):
        self.name = "water tank"
        self.pickup_string = "Pick up a water tank?"


class CoalStorageRoom:
    type = ITEM_COALSTORAGEROOM

    def __init__(self):
        self.name = "coal storage room"
        self.pickup_string = "Pick up a coal storage room?"


class Armour:
    type = ITEM_ARMOUR

    def __init__(self):
        self.name = "armour"
        self.pickup_string = "Pick up some armour?"


def coal_new(g, t, value):
    g.clayer[t.ty][t.tx] = 0
    s = isovid.Sprite(g.images["coal"], t.rect)
    s.frame = 0

    s.item = items.Coal(1.0)

    g.sprites.append(s)
    s.loop = coal_loop

    s.groups = g.string2groups("coal")
    s.agroups = g.string2groups("castle")
    s.hit = coal_hit

    from . import fire

    # s.effect = fire.Effect(64,1)


def coal_hit(g, s, a):
    # print "coal hit"
    pass


def coal_loop(g, s):
    pass
    # s.rect.clamp_ip(g.view)
    s.frame += 1


def cannon_tower_new(g, t, value):
    g.clayer[t.ty][t.tx] = 0
    s = isovid.Sprite(g.images["cannon_tower"], t.rect)

    s.item = items.Coal(1.0)

    g.sprites.append(s)
    s.loop = cannon_tower_loop

    # s.groups = g.string2groups('cannon_tower')
    # s.agroups = g.string2groups('castle')
    # s.hit = cannon_tower_hit


def cannon_tower_hit(g, s, a):
    pass


def cannon_tower_loop(g, s):
    pass


def water_new(g, t, value):
    g.clayer[t.ty][t.tx] = 0

    s = isovid.Sprite(g.images["water"], t.rect)
    g.sprites.append(s)
    # s.groups = g.string2groups('coal')
    # s.agroups = g.string2groups('castle')
    # s.hit = coal_hit

    s.frame = random.randrange(0, 32)
    s.loop = water_loop


def water_loop(g, s):
    # if s.frame%20 == 0:
    #    import explode
    #    s.effect = explode.Effect(512,128)
    #    pass

    s.frame += 1


def human_new(g, t, value):
    g.clayer[t.ty][t.tx] = 0

    s = isovid.Sprite(g.images["human"], t.rect)
    g.sprites.append(s)

    s.frame = random.randrange(0, 32)
    s.loop = human_loop


def human_loop(g, s):
    s.frame += 1


def wall_new(g, t, value):
    g.clayer[t.ty][t.tx] = 0

    s = isovid.Sprite(g.images["wall"], t.rect)
    g.sprites.append(s)

    s.frame = random.randrange(0, 32)
    s.loop = wall_loop


def wall_loop(g, s):
    s.frame += 1


def baddie_robot_new(g, t, value):
    g.clayer[t.ty][t.tx] = 0

    s = isovid.Sprite(g.images["baddie_robot"], t.rect)
    g.sprites.append(s)

    s.frame = random.randrange(0, 32)
    s.loop = baddie_robot_loop


def baddie_robot_loop(g, s):
    s.frame += 1


# castle
# human
# factory
# coal
# water
# cannon towers
# walls
# baddie robots. not castles, but smaller sleaker ones.



import pygame
from pygame.locals import *
import random

from .pgu import timer

rr = random.randrange


class Part:
    def __init__(self, pos):
        self.pos = pygame.Rect(pos[0], pos[1], 1, 1)
        self._pos = pygame.Rect(self.pos)
        self.frame = 0


SIZE = 16  # 32 is original


class Effect:
    def __init__(self, total, add):
        if not hasattr(Effect, "init"):
            Effect._init(self)

        self.total = total
        self._total = 0
        self.add = add

        self.parts = []

        self.frame = 0

    def _init(self):
        Effect.init = True
        Effect.colors = colors = []
        for r in range(0, 32):
            if r < 8:
                rr = r - 0
                c = (255, 255, 255 - rr * 8)
            elif r < 16:
                rr = r - 8
                c = (255, 255 - rr * 16, 255 - (rr + 8) * 8)
            else:
                rr = r - 16
                c = (255 - rr * 8, 128 - rr * 8, 128 - rr * 8)

            colors.append(c)

        Effect.images = images = []
        for v in range(0, 32):
            r = max(1, v * SIZE // 32)
            img = pygame.Surface((r * 2, r * 2)).convert()
            img.fill((255, 0, 255))
            img.set_colorkey((255, 0, 255))
            img.set_alpha(255 - v * 8)

            if v < 8:
                rr = v - 0
                c = (255, 255, 255 - rr * 8)
            elif v < 16:
                rr = v - 8
                c = (255, 255 - rr * 16, 255 - (rr + 8) * 8)
            else:
                rr = v - 16
                c = (255 - rr * 8, 128 - rr * 8, 128 - rr * 8)
            # print c
            pygame.draw.circle(img, c, (r, r), max(1, r // 2))
            # pygame.draw.circle(img,(255-r*8,255-r*8,255-r*8),(r,r),r)
            images.append(img)

    def paint(self, screen, origin):
        if self.frame > 16:
            return

        _screen = screen

        center = self.center
        sz = 240 * SIZE // 32
        screen = pygame.Surface((sz, sz)).convert()
        screen.fill((80, 0, 0))
        screen.set_colorkey((80, 0, 0))
        f = max(0, self.frame - 8)
        screen.set_alpha(255 - f * 32)
        adj = (sz // 2 - center[0], sz // 2 - center[1])

        center = self.center

        todo = [[] for n in range(0, 32)]

        for part in self.parts:
            p, _p = part.pos, part._pos
            xx, yy = abs(center[0] - p.x), abs(center[1] - p.y)
            # f = self.frame*2
            # r = ( abs((xx*xx+yy*yy) - (f*f)))/256
            r = (xx * xx + yy * yy) // (256 * SIZE // 32)
            # dx,dy = p.x-_p.x,p.y-_p.y
            # r = ((xx*xx+yy*yy)*(dx*dx+dy*dy))/16384
            if r >= 0 and r < 32:
                todo[r].append((_p, p))

        ox, oy = adj
        for v in range(31, -1, -1):
            img = Effect.images[v]
            c = Effect.colors[v]
            # v = 255-r*8
            # c = (v,v,v)
            # c = (255,255,255,v)
            r = img.get_width() // 2  # max(1,v*SIZE/32)
            sz = max(1, ((32 - v) // 3 + 1) * SIZE // 32)
            for _p, p in todo[v]:
                # pygame.draw.circle(screen,c,(p.x+pos[0],p.y+pos[1]),r+3)
                # screen.blit(img,(p.x-r+pos[0],p.y-r+pos[1]))

                dx, dy = p.x - _p.x, p.y - _p.y

                # sz = self.frame
                pygame.draw.line(
                    screen,
                    c,
                    (ox + _p.x - dx * 1, oy + _p.y - dy * 1),
                    (ox + p.x, oy + p.y),
                    sz,
                )
                screen.blit(img, (p.x - r + adj[0], p.y - r + adj[1]))

        _screen.blit(
            screen,
            (
                -origin[0] + center[0] - screen.get_width() // 2,
                -origin[1] + center[1] - screen.get_height() // 2,
            ),
        )

    def loop(self, pos):
        if self.frame > 20:
            return
        self.frame += 1

        self.center = pos

        parts = self.parts
        if self._total < self.total:
            for n in range(0, self.add):
                part = Part(
                    (
                        pos[0] + rr(-12 * SIZE // 32, 13 * SIZE // 32),
                        pos[1] + rr(-12 * SIZE // 32, 13 * SIZE // 32),
                    )
                )
                part._pos.x, part._pos.y = pos[0], pos[1]
                parts.append(part)
                self._total += 1
        l = len(parts)
        rm = []
        for part in self.parts:
            p, _p = part.pos, part._pos

            dx, dy = p.x - _p.x, p.y - _p.y
            _p.x, _p.y = p.x, p.y

            p.x += dx
            p.y += dy


if __name__ == "__main__":
    screen = pygame.display.set_mode((640, 480))

    # ss = [((rr(160,640),rr(160,480)),Effect(256,64)) for n in xrange(0,1)]

    ss = []

    t = timer.Timer(40)

    f = 0

    _quit = False
    while not _quit:
        for e in pygame.event.get():
            if e.type == QUIT:
                _quit = True
            if e.type == KEYDOWN and e.key == K_ESCAPE:
                _quit = True

        screen.fill((0, 0, 0))
        for pos, s in ss:
            s.loop(pos)
            s.paint(screen, (0, 0))
        pygame.display.flip()

        if f % 20 == 0:
            ss = [((rr(160, 480), rr(160, 320)), Effect(512, 128))]
        f += 1

        t.tick()




from setuptools import setup, find_packages

setup(
    name='zanthor',
    version="1.2.4a2",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.0",
        "Programming Language :: Python :: 3.1",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        'Topic :: Software Development :: Libraries :: pygame',
        'Topic :: Games/Entertainment :: Arcade',
        "Topic :: Games/Entertainment :: Real Time Strategy",
    ],
    license='GPL',
    author="Aaron, Phil, Rene, Tim",
    author_email="renesd@gmail.com",
    maintainer='Rene Dudfield',
    maintainer_email='renesd@gmail.com',
    description="Zanthor is a game where you play an evil robot castle which is powered by steam.  @zanthorgame #python #pygame",
    url="http://github.com/pygame/zanthor/",
    include_package_data=True,
    long_description='Zanthor is a game where you play an evil robot castle which is powered by steam.',
    package_dir={'zanthor': 'zanthor'},
    packages=find_packages(),
    install_requires=['pygame'],
    entry_points={
        'console_scripts': [
            'zanthor=zanthor.main:main',
        ],
    },
)



""" this file has lots of tweaky things in it.

Tweakable files (other) :
    sound_info.py - tweaking sounds.
    units.py - tweaking unit stats.

"""


import pygame

import os

_DATA_DIR = None


def data_dir(*args):
    global _DATA_DIR
    if _DATA_DIR is None:
        # _DATA_DIR = os.path.join(*(['zanthor', 'data'] + list(args)))
        _DATA_DIR = os.path.join("zanthor", "data")
        if not os.path.exists(_DATA_DIR):
            _DATA_DIR = os.path.join(os.path.split(__file__)[0], "data")
    return os.path.join(*([_DATA_DIR] + list(args)))


# the base size.
BSW, BSH = 640, 480
SW, SH = 320, 200
SW, SH = 800, 600
SW, SH = 1280, 800
SW, SH = 1024, 768
SW, SH = 640, 480

DX = float(SW) / BSW
DY = float(SH) / BSH

dxdy = (DX, DY)

FPS = 30

MIN_CANNON_PRESSURE = 6


def multr(dx_dy, r):
    a, b, c, d = tuple(r)
    dx, dy = dxdy
    return pygame.Rect(a * dx, b * dy, c * dx, d * dy)


# S_VIEW = pygame.Rect(0,0,480,480)


S_STATUS = multr(dxdy, (480, 320, 400, 160))

S_ITEMS = multr(dxdy, (480, 160, 160, 200))
S_STATUS = multr(dxdy, (480, 360, 160, 120))


# game view area.
S_VIEW = multr(dxdy, (76, 0, 380, 340))


# left side, full status area.
S_STATUS = multr(dxdy, (0, 0, 76, 340))


# left side status boxes.
# S_HEALTH = pygame.Rect(10,30,60,70)
S_HEALTH = multr(dxdy, (10, 4, 60, 96))
S_COAL = multr(dxdy, (10, 100, 60, 70))
S_WATER = multr(dxdy, (10, 190, 60, 70))
S_STEAM = multr(dxdy, (10, 260, 60, 70))

S_PEASANTS_REMAINING = multr(dxdy, (487, 24, 122, 114))
S_CANNON_PRESSURE = multr(dxdy, (76, 0, 5, 340))


# top right,  castle robot illustration area.
S_ROBOT = multr(dxdy, (456, 0, 184, 163))


# middle right, items area.
S_ITEMS = multr(dxdy, (456, 163, 184, 177))


S_ITEMS_ENGINE = multr(dxdy, (480, 160, 160, 160))
S_ITEMS_COALSTORAGEROOM = multr(dxdy, (480, 160, 160, 160))
S_ITEMS_WATERTANK = multr(dxdy, (480, 160, 160, 160))
S_ITEMS_CANNON = multr(dxdy, (480, 160, 160, 160))
S_ITEMS_ARMOUR = multr(dxdy, (480, 160, 160, 160))


# bottom right,  button area.
S_BUTTONS = pygame.Rect(470, 340, 170, 140)

S_BUTTONS_SAVE = pygame.Rect(490, 360, 60, 40)
S_BUTTONS_LOAD = pygame.Rect(490, 420, 60, 40)
S_BUTTONS_QUIT = pygame.Rect(560, 360, 60, 40)
S_BUTTONS_NEWS = pygame.Rect(560, 420, 60, 40)

S_BOTTOM = pygame.Rect(0, 340, 640, 140)

(
    S_BUTTONS,
    S_BUTTONS_SAVE,
    S_BUTTONS_LOAD,
    S_BUTTONS_QUIT,
    S_BUTTONS_NEWS,
    S_BOTTOM,
) = [
    multr(dxdy, x)
    for x in [
        S_BUTTONS,
        S_BUTTONS_SAVE,
        S_BUTTONS_LOAD,
        S_BUTTONS_QUIT,
        S_BUTTONS_NEWS,
        S_BOTTOM,
    ]
]


S_BOTTOM_BUTTONS = multr(dxdy, (470, 0, 170, 140))
S_BOTTOM_MESSAGES = multr(dxdy, (95, 25, 355, 110))
# bottom middle message area.
S_MESSAGES = multr(dxdy, (95, 365, 355, 110))

# S_NEXT_LEVEL_MESSAGES = pygame.Rect(20,165,380,410)
S_NEXT_LEVEL_MESSAGES = multr(dxdy, (20, 20, 500, 410))


S_ = pygame.Rect(480, 360, 160, 120)


UPGRADE_FUN = 1

CACHE_USE_LEVEL_CACHE = 1

# if you are modifying just the tiles or just the levels set one of these to 0.
#  These only check the levels,
#    or the tiles to see if its older than the cache.
CACHE_CHECK_TILES = 1
CACHE_CHECK_LEVEL = 1


# joystick buttons
JOY_FIRE_BUTTON = 1
JOY_PICKUP_BUTTON = 2

# should things be auto picked up or not?
AUTO_PICKUP = 1

# should holes that the castle hits the ground be
# solid tiles
# 0 = no, 1 = yes, 2= big holes
HOLES_ARE_TILES = 0


# NOTE: mouselook seems to not work well on some systems...
#       can toggle it with the m key in game.
DISABLE_MOUSE_LOOK = 0


FLOCKING_FLUCT = 1


def get_mouse_info():
    if DISABLE_MOUSE_LOOK:
        SCROLL_MOUSE = 0  # speed of mouse scroll
        SCROLL_AUTO = 9  # speed of auto-return scroll
        SCROLL_BORDER = 70  # size of scrolling border during auto-scroll
    else:
        SCROLL_MOUSE = 12  # speed of mouse scroll
        SCROLL_AUTO = 9  # speed of auto-return scroll
        SCROLL_BORDER = 160  # size of scrolling border during auto-scroll

    return [SCROLL_MOUSE, SCROLL_AUTO, SCROLL_BORDER]



"""Square tile based engine."""

from pgu.vid import *
import pygame


class Tilevid(Vid):
    """Based on [[vid]] -- see for reference."""

    def paint(self, s):
        sw, sh = s.get_width(), s.get_height()
        self.view.w, self.view.h = sw, sh

        tiles = self.tiles
        tw, th = tiles[0].image.get_width(), tiles[0].image.get_height()
        w, h = self.size

        if self.bounds != None:
            self.view.clamp_ip(self.bounds)

        ox, oy = self.view.x, self.view.y
        tlayer = self.tlayer
        blayer = self.blayer
        alayer = self.alayer
        sprites = self.sprites

        blit = s.blit
        yy = -(self.view.y % th)
        my = (oy + sh) / th
        if (oy + sh) % th:
            my += 1

        if blayer != None:
            for y in range(oy / th, my):
                if y >= 0 and y < h:
                    trow = tlayer[y]
                    brow = blayer[y]
                    arow = alayer[y]
                    xx = -(self.view.x % tw)
                    mx = (ox + sw) / tw
                    # if (ox+sh)%tw: mx += 1
                    for x in range(ox / tw, mx + 1):
                        if x >= 0 and x < w:
                            blit(tiles[brow[x]].image, (xx, yy))
                            blit(tiles[trow[x]].image, (xx, yy))
                            arow[x] = 0
                        xx += tw
                yy += th
        else:
            for y in range(oy / th, my):
                if y >= 0 and y < h:
                    trow = tlayer[y]
                    arow = alayer[y]
                    xx = -(self.view.x % tw)
                    mx = (ox + sw) / tw
                    # if (ox+sh)%tw: mx += 1
                    for x in range(ox / tw, mx + 1):
                        if x >= 0 and x < w:
                            blit(tiles[trow[x]].image, (xx, yy))
                            arow[x] = 0
                        xx += tw
                yy += th

        for s in sprites:
            s.irect.x = s.rect.x - s.shape.x
            s.irect.y = s.rect.y - s.shape.y
            blit(s.image, (s.irect.x - ox, s.irect.y - oy))
            s.updated = 0
            s._irect = Rect(s.irect)
            # s._rect = Rect(s.rect)

        self.updates = []
        self._view = pygame.Rect(self.view)
        return [Rect(0, 0, sw, sh)]

    def update(self, s):
        sw, sh = s.get_width(), s.get_height()
        self.view.w, self.view.h = sw, sh

        if self.bounds != None:
            self.view.clamp_ip(self.bounds)
        if self.view.x != self._view.x or self.view.y != self._view.y:
            return self.paint(s)

        ox, oy = self.view.x, self.view.y
        sw, sh = s.get_width(), s.get_height()
        w, h = self.size
        tlayer = self.tlayer
        blayer = self.blayer
        alayer = self.alayer
        tiles = self.tiles
        tw, th = tiles[0].image.get_width(), tiles[0].image.get_height()
        sprites = self.sprites
        blit = s.blit

        us = []

        # mark places where sprites have moved, or been removed

        ss = self.sprites.removed
        self.sprites.removed = []
        ss.extend(sprites)
        for s in ss:
            # figure out what has been updated.
            s.irect.x = s.rect.x - s.shape.x
            s.irect.y = s.rect.y - s.shape.y
            if (
                s.irect.x != s._irect.x
                or s.irect.y != s._irect.y
                or s.image != s._image
            ):
                # w,h can be skipped, image covers that...
                s.updated = 1
            if s.updated:
                r = s._irect
                y = max(0, r.y / th)
                yy = min(h, r.bottom / th + 1)
                while y < yy:
                    x = max(0, r.x / tw)
                    xx = min(w, r.right / tw + 1)
                    while x < xx:
                        if alayer[y][x] == 0:
                            self.updates.append((x, y))
                        alayer[y][x] = 1
                        x += 1
                    y += 1

                r = s.irect
                y = max(0, r.y / th)
                yy = min(h, r.bottom / th + 1)
                while y < yy:
                    x = r.x / tw
                    xx = min(w, r.right / tw + 1)
                    while x < xx:
                        if alayer[y][x] == 0:
                            alayer[y][x] = 2
                            self.updates.append((x, y))
                        x += 1
                    y += 1

        # mark sprites that are not being updated that need to be updated because
        # they are being overwritte by sprites / tiles
        for s in sprites:
            if s.updated == 0:
                r = s.irect
                y = max(0, r.y / th)
                yy = min(h, r.bottom / th + 1)
                while y < yy:
                    x = max(0, r.x / tw)
                    xx = min(w, r.right / tw + 1)
                    while x < xx:
                        if alayer[y][x] == 1:
                            s.updated = 1
                        x += 1
                    y += 1

        for u in self.updates:
            x, y = u
            xx, yy = x * tw - ox, y * th - oy
            if alayer[y][x] == 1:
                if blayer != None:
                    blit(tiles[blayer[y][x]].image, (xx, yy))
                blit(tiles[tlayer[y][x]].image, (xx, yy))
            alayer[y][x] = 0
            us.append(Rect(xx, yy, tw, th))

        for s in sprites:
            if s.updated:
                blit(s.image, (s.irect.x - ox, s.irect.y - oy))
                s.updated = 0
                s._irect = Rect(s.irect)
                s._image = s.image

        self.updates = []
        return us

    def view_to_tile(self, pos):
        x, y = pos
        tiles = self.tiles
        tw, th = tiles[0].image.get_width(), tiles[0].image.get_height()
        return x / tw, y / th

    def tile_to_view(self, pos):
        x, y = pos
        tiles = self.tiles
        tw, th = tiles[0].image.get_width(), tiles[0].image.get_height()
        x, y = x * tw, y * th
        return x, y

    def screen_to_tile(self, pos):
        x, y = pos
        x, y = x + self.view.x, y + self.view.y
        return self.view_to_tile((x, y))

    def tile_to_screen(self, pos):
        x, y = pos
        x, y = self.tile_to_view(pos)
        x, y = x - self.view.x, y - self.view.y
        return x, y


# vim: set filetype=python sts=4 sw=4 noet si :



"""
"""
import pygame
from pygame.locals import *

from .const import *
from . import widget


class Keysym(widget.Widget):
    """A keysym input.

    <p>This widget records the keysym of the key pressed while this widget is in focus.</p>

    <pre>Keysym(value=None)</pre>

    <dl>
    <dt>value<dd>initial keysym, see <a href="http://www.pygame.org/docs/ref/key.html">pygame keysyms</a> </dl>

    <strong>Example</strong>
    <code>
    w = Input(value=pygame.locals.K_g)

    w = Input(pygame.locals.K_g)

    w = Input()
    </code>

    """

    def __init__(self, value=None, **params):
        params.setdefault("cls", "keysym")
        widget.Widget.__init__(self, **params)
        self.value = value

        self.font = self.style.font
        w, h = self.font.size("Right Super")  # "Right Shift")
        self.style.width, self.style.height = w, h
        # self.rect.w=w+self.style.padding_left+self.style.padding_right
        # self.rect.h=h+self.style.padding_top+self.style.padding_bottom

    def event(self, e):
        if e.type == FOCUS or e.type == BLUR:
            self.repaint()
        elif e.type == KEYDOWN:
            if e.key != K_TAB:
                self.value = e.key
                self.repaint()
                self.send(CHANGE)
            next(self)
        self.pcls = ""
        if self.container.myfocus is self:
            self.pcls = "focus"

    def paint(self, s):
        r = pygame.rect.Rect(0, 0, self.rect.w, self.rect.h)
        # render_box(s,self.style.background,r)
        if self.value == None:
            return
        name = ""
        for p in pygame.key.name(self.value).split():
            name += p.capitalize() + " "
        # r.x = self.style.padding_left;
        # r.y = self.style.padding_bottom;
        s.blit(self.style.font.render(name, 1, self.style.color), r)

    def __setattr__(self, k, v):
        if k == "value" and v != None:
            v = int(v)
        _v = self.__dict__.get(k, NOATTR)
        self.__dict__[k] = v
        if k == "value" and _v != NOATTR and _v != v:
            self.send(CHANGE)
            self.repaint()



"""
"""
import os

from .const import *
from . import table, area
from . import basic, input, button


class Dialog(table.Table):
    """A dialog window with a title bar and an "close" button on the bar.

    <pre>Dialog(title,main)</pre>

    <dl>
    <dt>title<dd>title widget, usually a label
    <dt>main<dd>main widget, usually a container
    </dl>

    <strong>Example</strong>
    <code>
    title = gui.Label("My Title")
    main = gui.Container()
    #add stuff to the container...

    d = gui.Dialog(title,main)
    d.open()
    </code>
    """

    def __init__(self, title, main, **params):
        params.setdefault("cls", "dialog")
        table.Table.__init__(self, **params)

        self.tr()
        self.td(title, align=-1, cls=self.cls + ".bar")
        clos = button.Icon(self.cls + ".bar.close")
        clos.connect(CLICK, self.close, None)
        self.td(clos, align=1, cls=self.cls + ".bar")

        self.tr()
        self.td(main, colspan=2, cls=self.cls + ".main")


#         self.tr()
#
#
#         t = table.Table(cls=self.cls+".bar")
#         t.tr()
#         t.td(title)
#         clos = button.Icon(self.cls+".bar.close")
#         t.td(clos,align=1)
#         clos.connect(CLICK,self.close,None)
#         self.add(t,0,0)
#
#         main.rect.w,main.rect.h = main.resize()
#         clos.rect.w,clos.rect.h = clos.resize()
#         title.container.style.width = main.rect.w - clos.rect.w
#
#         self.tr()
#         self.td(main,cls=self.cls+".main")
#


class FileDialog(Dialog):
    """A file picker dialog window.

    <pre>FileDialog()</pre>
    <p>Some optional parameters:</p>
    <dl>
    <dt>title_txt<dd>title text
    <dt>button_txt<dd>button text
    <dt>path<dd>initial path
    </dl>
    """

    def __init__(
        self, title_txt="File Browser", button_txt="Okay", cls="filedialog", path=None
    ):
        if not path:
            self.curdir = os.getcwd()
        else:
            self.curdir = path
        from . import app

        self.dir_img = basic.Image(app.App.app.theme.get(cls + ".folder", "", "image"))
        td_style = {
            "padding_left": 4,
            "padding_right": 4,
            "padding_top": 2,
            "padding_bottom": 2,
        }
        self.title = basic.Label(title_txt, cls=cls + ".title.label")
        self.body = table.Table()
        self.list = area.List(width=350, height=150)
        self.input_dir = input.Input(cls=cls + ".input")
        self.input_file = input.Input(cls=cls + ".input")
        self._list_dir_()
        self.button_ok = button.Button(button_txt)
        self.body.tr()
        self.body.td(basic.Label("Path", cls=cls + ".label"), style=td_style, align=-1)
        self.body.td(self.input_dir, style=td_style)
        self.body.tr()
        self.body.td(basic.Label("File", cls=cls + ".label"), style=td_style, align=-1)
        self.body.td(self.input_file, style=td_style)
        self.body.td(self.button_ok, style=td_style)
        self.body.tr()
        self.body.td(self.list, colspan=3, style=td_style)
        self.list.connect(CHANGE, self._item_select_changed_, None)
        self.button_ok.connect(CLICK, self._button_okay_clicked_, None)
        self.value = None
        Dialog.__init__(self, self.title, self.body)

    def _list_dir_(self):
        self.input_dir.value = self.curdir
        self.input_dir.pos = len(self.curdir)
        self.input_dir.vpos = 0
        dirs = []
        files = []
        try:
            for i in os.listdir(self.curdir):
                if os.path.isdir(os.path.join(self.curdir, i)):
                    dirs.append(i)
                else:
                    files.append(i)
        except:
            self.input_file.value = "Opps! no access"
        if ".." not in dirs:
            dirs.append("..")
        dirs.sort()
        files.sort()
        for i in dirs:
            # item = ListItem(image=self.dir_img, text=i, value=i)
            self.list.add(i, image=self.dir_img, value=i)
        for i in files:
            # item = ListItem(image=None, text=i, value=i)
            self.list.add(i, value=i)
        # self.list.resize()
        self.list.set_vertical_scroll(0)
        # self.list.repaintall()

    def _item_select_changed_(self, arg):
        self.input_file.value = self.list.value
        fname = os.path.abspath(os.path.join(self.curdir, self.input_file.value))
        if os.path.isdir(fname):
            self.input_file.value = ""
            self.curdir = fname
            self.list.clear()
            self._list_dir_()

    def _button_okay_clicked_(self, arg):
        if self.input_dir.value != self.curdir:
            if os.path.isdir(self.input_dir.value):
                self.input_file.value = ""
                self.curdir = os.path.abspath(self.input_dir.value)
                self.list.clear()
                self._list_dir_()
        else:
            self.value = os.path.join(self.curdir, self.input_file.value)
            self.send(CHANGE)
            self.close()



import pygame
from pygame.locals import *
import random

from .pgu import timer

rr = random.randrange


class Part:
    def __init__(self, pos, z):
        self.pos = pygame.Rect(pos[0], pos[1], 1, 1)
        self._pos = pygame.Rect(self.pos)
        self.frame = 0
        self.z = z


NUM_BITS = 10


class Effect:
    def __init__(self, total, add, region=1, respawn=1, color=(255, 255, 255)):
        # if not hasattr(Effect,'images'):
        #    Effect._init(self)

        self._init(color)

        self.total = total
        self.add = add

        self.parts = []
        self.region = region
        self.respawn = respawn

        self.frame = 0
        self.frames = 0

    def _init(self, color):
        if not hasattr(Effect, "data"):
            Effect.data = {}
        data = Effect.data
        k = "images.%s" % str(color)
        if k not in data:
            images = data[k] = []
            for r in range(0, NUM_BITS):
                img = pygame.Surface((r * 2, r * 2)).convert()
                img.fill((255, 0, 255))
                img.set_colorkey((255, 0, 255))
                img.set_alpha(255 - r * (8 * 32 / NUM_BITS))
                pygame.draw.circle(img, color, (r, r), r)
                # pygame.draw.circle(img,(255-r*8,255-r*8,255-r*8),(r,r),r)
                images.append(img)
        self.images = data[k]

    def zpaint(self):  # ,origin):
        ret = []

        if 1 or not self.frames % 2:
            todo = [[] for n in range(0, NUM_BITS)]
            for part in self.parts:
                p = part.pos
                r = part.frame
                if r >= 0 and r < NUM_BITS:
                    todo[r].append(part)
            for r in range(NUM_BITS - 1, -1, -1):
                img = self.images[r]
                v = 255 - r * 8
                # c = (v,v,v)
                # c = (255,255,255,v)
                for part in todo[r]:
                    p = part.pos
                    # pygame.draw.circle(screen,c,(p.x+pos[0],p.y+pos[1]),r+3)
                    # screen.blit(img,(-origin[0]+p.x-r,-origin[1]+p.y-r))
                    ret.append((part.z, img, (p.x - r, p.y - r)))

        self.frames += 1

        return ret

    def zloop(self, pos, z):
        parts = self.parts
        if len(parts) < self.total:
            region = self.region
            for n in range(0, self.add):
                part = Part(
                    (pos[0] + rr(-region, region), pos[1] + rr(-region, region)), z
                )
                parts.append(part)
        for part in parts:
            p, _p = part.pos, part._pos

            dx, dy = p.x - _p.x, p.y - _p.y
            _p.x, _p.y = p.x, p.y

            p.x += dx / 3 + rr(-4, 2)
            p.y += dy / 4 + rr(-6, 2)

            # r = (abs(p.x)+abs(p.y))/8
            # if r > NUM_BITS:
            part.frame += 1

            if part.frame >= NUM_BITS and self.respawn:
                p.x, p.y = pos[0] + rr(-2, 2), pos[1] + rr(-2, 0)
                _p.x, _p.y = p.x, p.y
                part.z = z
                part.frame = 0

    def paint(self, screen, origin):
        for z, img, pos in self.zpaint():
            screen.blit(img, (-origin[0] + pos[0], -origin[1] + pos[1]))

    def loop(self, pos):
        return self.zloop(pos, 0)


if __name__ == "__main__":
    screen = pygame.display.set_mode((640, 480))

    ss = [((rr(160, 640), rr(160, 480)), Effect(NUM_BITS, 1)) for n in range(0, 8)]

    # t = timer.Timer(40)
    t = timer.Speedometer()
    pygame.time.get_ticks()

    _quit = False
    while not _quit:
        for e in pygame.event.get():
            if e.type == QUIT:
                _quit = True
            if e.type == KEYDOWN and e.key == K_ESCAPE:
                _quit = True

        screen.fill((0, 0, 0))
        for pos, s in ss:
            s.loop(pos)
            s.paint(screen, (0, 0))
        pygame.display.flip()

        r = t.tick()
        if r:
            print(r)



# this file gets all the dinky little generic states

import pygame
from pygame.locals import *

import os

from .pgu import engine
from .pgu import text

from .const import *
from . import util
import random

pgu_text_write = text.write

class Title(engine.State):
    def init(self):
        self.menu = [
            ("Start", "start"),
            # ('Load','load'),
            # ('Save','save'),
            ("Help", Help),
            ("Credits", Credits),
            ("Quit", engine.Quit),
        ]
        self.cur = 0

        if pygame.mixer:
            pygame.mixer.music.load(data_dir("intro", "zanthor.ogg"))
            pygame.mixer.music.play(-1)

        self.font_title = []
        for size in (160, 165, 170, 175, 180):
            self.font_title.append(
                pygame.font.Font(data_dir("intro", "WALSHES.TTF"), size)
            )
        self.font_main = pygame.font.Font(data_dir("menu", "vinque.ttf"), 32)

        self.rects = []
        pygame.mouse.set_visible(True)
        self.bkgr = pygame.image.load(data_dir("intro", "mybkgr.png"))

        self.frame = 0

    def paint(self, screen):
        # print self.game.backup_castle_stats
        screen.fill((255, 0, 0))

        img = self.bkgr.subsurface((404, 121, 236, 359))
        img = pygame.transform.rotozoom(
            img, random.randrange(-15, 15), random.randrange(90, 115) / 100.0
        )
        x, y = 380 - img.get_width() / 2 + random.randrange(
            -24, 24
        ), 290 - img.get_height() / 2 + random.randrange(-12, 12)
        screen.blit(img, (x, y))

        fnt = random.choice(self.font_title)
        text = "Zanthor"
        fg, bg, b = (255, 255, 255), (0, 0, 0), 4
        if self.frame % 8 < 4:
            fg = (255, 0x55, 0x55)

        img = fnt.render(text, 1, bg)
        x, y = (SW - img.get_width()) / 2 + random.randrange(
            -8, 8
        ), 100 - img.get_height() / 2 + random.randrange(-8, 8)
        for dx, dy in [
            (-1, 0),
            (0, 1),
            (1, 0),
            (0, -1),
            (-1, -1),
            (-1, 1),
            (1, -1),
            (1, 1),
        ]:
            screen.blit(img, (x + dx * b, y + dy * b))
        img = fnt.render(text, 1, fg)
        screen.blit(img, (x, y))

        fnt = self.font_main
        x, y, n = 48, 220, 0
        self.rects = []

        for text, action in self.menu:
            c = (0xAA, 0xAA, 0xAA)
            if n == self.cur:
                c = (255, 255, 255)
            bg = (0, 0, 0)
            img = fnt.render(text, 1, bg)
            b = 2
            for dx, dy in [(-1, 0), (0, 1), (1, 0), (0, -1)]:
                screen.blit(img, (x + dx * b, y + dy * b))

            img = fnt.render(text, 1, c)
            screen.blit(img, (x, y))
            self.rects.append((n, pygame.Rect(x, y, img.get_width(), img.get_height())))
            y += 40
            n += 1
        # print self.cur,n

        pygame.display.flip()

    def loop(self):
        if self.frame % 3 == 0:
            self.repaint()
        self.frame += 1

    def event(self, e):
        print('title', e, e.type == KEYDOWN, e.type, KEYDOWN, K_UP, K_DOWN)
        if e.type == MOUSEMOTION:
            for n, r in self.rects:
                if r.collidepoint(e.pos):
                    self.cur = n
                    # self.repaint()

        elif e.type == JOYAXISMOTION:
            if e.axis == 1:
                if round(e.value) < 0:
                    self.cur = (self.cur - 1 + len(self.menu)) % len(self.menu)
                elif round(e.value) > 0:
                    self.cur = (self.cur + 1 + len(self.menu)) % len(self.menu)

        elif e.type == KEYDOWN:
            print('keydown')
            if e.key == K_UP:
                print('UP', self.cur)
                self.cur = (self.cur - 1 + len(self.menu)) % len(self.menu)
                print('UP after', self.cur)
                # self.repaint()
            if e.key == K_DOWN:
                print('DOWN', self.cur)
                self.cur = (self.cur + 1 + len(self.menu)) % len(self.menu)
                print('DOWN after', self.cur)
                # self.repaint()

            
        if (e.type == KEYDOWN and e.key == K_RETURN) or e.type in [
            MOUSEBUTTONDOWN,
            JOYBUTTONDOWN,
        ]:
            text, action = self.menu[self.cur]
            if action == "start":
                if pygame.mixer:
                    pygame.mixer.music.stop()
                g = self.game
                g.data_reset()
                from . import menu

                return menu.Menu(self.game)
            else:
                return action(self.game)


# etc...


class Help(engine.State):
    def paint(self, screen):
        screen.blit(pygame.image.load(data_dir("intro", "mybkgr.png")), (0, 0))

        text = [
            "help",
            "",
            "press and hold right mouse",
            "to build up pressure.",
            "release to shoot.",
            "",
            "collect coal and water",
            "to keep steam up.",
            "",
            "arrow keys, a/w/s/d also",
            "work.  f/space to fire.",
            "",
            "you are ZANTHOR!",
        ]

        fnt = pygame.font.Font(data_dir("menu", "vinque.ttf"), 24)
        x, y = 48, 20
        for line in text:
            img = fnt.render(line, 0, (0, 0, 0))
            # x = (SW-img.get_width())/2
            # screen.blit(img,(x,y))
            pgu_text_write(screen, fnt, (x, y), (255, 255, 255), line, 1)
            y += 32
        pygame.display.flip()

    def event(self, e):
        if e.type == KEYDOWN or e.type == MOUSEBUTTONDOWN:
            return Title(self.game)


class Credits(engine.State):
    def paint(self, screen):
        screen.fill((255, 0, 0))

        screen.blit(pygame.image.load(data_dir("intro", "mybkgr.png")), (0, 0))
        text = [
            "credits",
            "",
            "aka - graphics",
            "illume - code, etc",
            "philhassey - code, etc",
            "TimInge - music, sfx",
            "",
        ]

        fnt = pygame.font.Font(data_dir("menu", "vinque.ttf"), 32)
        x, y = 60, 20
        for line in text:
            img = fnt.render(line, 0, (0, 0, 0))
            x = (460 - img.get_width()) / 2
            # screen.blit(img,(x,y))
            pgu_text_write(screen, fnt, (x, y), (255, 255, 255), line, 1)
            y += 48
        pygame.display.flip()

    def event(self, e):
        if e.type == KEYDOWN or e.type == MOUSEBUTTONDOWN:
            return Title(self.game)



"""
"""

from pygame.locals import *

from .const import *
from . import widget, surface
from . import basic


class _button(widget.Widget):
    def __init__(self, **params):
        widget.Widget.__init__(self, **params)
        self.state = 0

    def event(self, e):
        if e.type == ENTER:
            self.repaint()
        elif e.type == EXIT:
            self.repaint()
        elif e.type == FOCUS:
            self.repaint()
        elif e.type == BLUR:
            self.repaint()
        elif e.type == KEYDOWN:
            if e.key == K_SPACE:
                self.state = 1
                self.repaint()
            elif e.key == K_TAB:
                next(self)
        elif e.type == MOUSEBUTTONDOWN:
            self.state = 1
            self.repaint()
        elif e.type == KEYUP:
            if self.state == 1:
                sub = pygame.event.Event(CLICK, {})
                self.send(sub.type, sub)

            self.state = 0
            self.repaint()
        elif e.type == MOUSEBUTTONUP:
            self.state = 0
            self.repaint()
        elif e.type == CLICK:
            self.click()

        self.pcls = ""
        if self.state == 0 and self.container.myhover is self:
            self.pcls = "hover"
        if self.state == 1 and self.container.myhover is self:
            self.pcls = "down"

    def click(self):
        pass


class Button(_button):
    """A button, buttons can be clicked, they are usually used to set up callbacks.

    <pre>Button(value=None)</pre>

    <dl>
    <dt>value<dd>either a widget or a string
    </dl>

    <strong>Example</strong>
    <code>
    w = gui.Button("Click Me")
    w.connect(gui.CLICK,fnc,value)
    </code>
    """

    def __init__(self, value=None, **params):
        params.setdefault("cls", "button")
        _button.__init__(self, **params)
        self.value = value

    def __setattr__(self, k, v):
        if k == "value" and type(v) == str:
            v = basic.Label(v, cls=self.cls + ".label")
        _v = self.__dict__.get(k, NOATTR)
        self.__dict__[k] = v
        if k == "value" and v != None:
            pass

        if k == "value" and _v != NOATTR and _v != None and _v != v:
            self.send(CHANGE)
            self.chsize()

    def resize(self, width=None, height=None):
        self.value.rect.x, self.value.rect.y = 0, 0
        self.value.rect.w, self.value.rect.h = self.value.resize(width, height)
        return self.value.rect.w, self.value.rect.h

    #
    #         self.value._resize()
    #         self.rect.w,self.rect.h = self.value.rect_margin.w,self.value.rect_margin.h
    #
    #         if self.style.width: self.rect.w = max(self.rect.w,self.style.width)
    #         if self.style.height: self.rect.w = max(self.rect.w,self.style.height)
    #
    #         xt,xr,xb,xl = self.value.getspacing()
    #
    #         self.value._resize(self.rect.w-(xl+xr),self.rect.h-(xt+xb))
    #
    def paint(self, s):
        self.value.paint(surface.subsurface(s, self.value.rect))


class Switch(_button):
    """A switch can have two states, True or False.

    <pre>Switch(value=False)</pre>

    <dl>
    <dt>value<dd>initial value, (True, False)
    </dl>

    <strong>Example</strong>
    <code>
    w = gui.Switch(True)
    w.connect(gui.CHANGE,fnc,value)
    </code>
    """

    def __init__(self, value=False, **params):
        params.setdefault("cls", "switch")
        _button.__init__(self, **params)
        self.value = value

        img = self.style.off
        self.style.width = img.get_width()
        self.style.height = img.get_height()

    def paint(self, s):
        # self.pcls = ""
        # if self.container.myhover is self: self.pcls = "hover"
        if self.value:
            img = self.style.on
        else:
            img = self.style.off
        s.blit(img, (0, 0))

    def __setattr__(self, k, v):
        _v = self.__dict__.get(k, NOATTR)
        self.__dict__[k] = v
        if k == "value" and _v != NOATTR and _v != v:
            self.send(CHANGE)
            self.repaint()

    def click(self):
        self.value = not self.value


class Checkbox(_button):
    """Within a Group of Checkbox widgets several may be selected at a time.

    <pre>Checkbox(group,value=None)</pre>

    <dl>
    <dt>group<dd>a gui.Group for the Checkbox to belong to
    <dt>value<dd>the value
    </dl>

    <strong>Example</strong>
    <code>
    g = gui.Group(name='colors',value=['r','b'])

    t = gui.Table()
    t.tr()
    t.td(gui.Label('Red'))
    t.td(gui.Checkbox(g,'r'))
    t.tr()
    t.td(gui.Label('Green'))
    t.td(gui.Checkbox(g,'g'))
    t.tr()
    t.td(gui.Label('Blue'))
    t.td(gui.Checkbox(g,'b'))
    </code>
    """

    def __init__(self, group, value=None, **params):
        params.setdefault("cls", "checkbox")
        _button.__init__(self, **params)
        self.group = group
        self.group.add(self)
        if self.group.value == None:
            self.group.value = []
        self.value = value

        img = self.style.off
        self.style.width = img.get_width()
        self.style.height = img.get_height()

    def paint(self, s):
        # self.pcls = ""
        # if self.container.myhover is self: self.pcls = "hover"
        if self.value in self.group.value:
            img = self.style.on
        else:
            img = self.style.off

        s.blit(img, (0, 0))

    def click(self):
        if self.value in self.group.value:
            self.group.value.remove(self.value)
        else:
            self.group.value.append(self.value)
        self.group._change()


class Radio(_button):
    """Within a Group of Radio widgets only one may be selected at a time.

    <pre>Radio(group,value=None)</pre>

    <dl>
    <dt>group<dd>a gui.Group for the Radio to belong to
    <dt>value<dd>the value
    </dl>

    <strong>Example</strong>
    <code>
    g = gui.Group(name='colors',value='g')

    t = gui.Table()
    t.tr()
    t.td(gui.Label('Red'))
    t.td(gui.Radio(g,'r'))
    t.tr()
    t.td(gui.Label('Green'))
    t.td(gui.Radio(g,'g'))
    t.tr()
    t.td(gui.Label('Blue'))
    t.td(gui.Radio(g,'b'))
    </code>
    """

    def __init__(self, group=None, value=None, **params):
        params.setdefault("cls", "radio")
        _button.__init__(self, **params)
        self.group = group
        self.group.add(self)
        self.value = value

        img = self.style.off
        self.style.width = img.get_width()
        self.style.height = img.get_height()

    def paint(self, s):
        # self.pcls = ""
        # if self.container.myhover is self: self.pcls = "hover"
        if self.group.value == self.value:
            img = self.style.on
        else:
            img = self.style.off
        s.blit(img, (0, 0))

    def click(self):
        self.group.value = self.value


class Tool(_button):
    """Within a Group of Tool widgets only one may be selected at a time.

    <pre>Tool(group,widget=None,value=None)</pre>

    <dl>
    <dt>group<dd>a gui.Group for the Tool to belong to
    <dt>widget<dd>a widget to appear on the Tool (similar to a Button)
    <dt>value<dd>the value
    </dl>

    <strong>Example</strong>
    <code>
    g = gui.Group(name='colors',value='g')

    t = gui.Table()
    t.tr()
    t.td(gui.Tool(g,'Red','r'))
    t.tr()
    t.td(gui.Tool(g,'Green','g'))
    t.tr()
    t.td(gui.Tool(g,'Blue','b'))
    </code>
    """

    def __init__(
        self, group, widget=None, value=None, **params
    ):  # TODO widget= could conflict with module widget
        params.setdefault("cls", "tool")
        _button.__init__(self, **params)
        self.group = group
        self.group.add(self)
        self.value = value

        if widget:
            self.setwidget(widget)

        if self.group.value == self.value:
            self.pcls = "down"

    def setwidget(self, w):
        self.widget = w

    def resize(self, width=None, height=None):
        self.widget.rect.w, self.widget.rect.h = self.widget.resize()
        # self.widget._resize()
        # self.rect.w,self.rect.h = self.widget.rect_margin.w,self.widget.rect_margin.h

        return self.widget.rect.w, self.widget.rect.h

    def event(self, e):
        _button.event(self, e)
        if self.group.value == self.value:
            self.pcls = "down"

    def paint(self, s):
        if self.group.value == self.value:
            self.pcls = "down"
        self.widget.paint(surface.subsurface(s, self.widget.rect))

    def click(self):
        self.group.value = self.value
        for w in self.group.widgets:
            if w != self:
                w.pcls = ""


class Icon(_button):
    """TODO - might be deprecated"""

    def __init__(self, cls, **params):
        _button.__init__(self, **params)
        self.cls = cls
        self.pcls = ""
        s = self.style.image
        self.style.width = s.get_width()
        self.style.height = s.get_height()
        self.state = 0

    def paint(self, s):
        # self.pcls = ""
        # if self.state == 0 and hasattr(self.container,'myhover') and self.container.myhover is self: self.pcls = "hover"
        # if self.state == 1 and hasattr(self.container,'myhover') and self.container.myhover is self: self.pcls = "down"
        s.blit(self.style.image, (0, 0))
